# This file is generated from pydcs_export.lua

from . import unittype


class Fortification:

    class Bunker_2(unittype.VehicleType):
        id = "Bunker"
        name = "Bunker 2"

    class Bunker_1(unittype.VehicleType):
        id = "Sandbox"
        name = "Bunker 1"

    class Barracks_armed(unittype.VehicleType):
        id = "house1arm"
        name = "Barracks armed"

    class Watch_tower_armed(unittype.VehicleType):
        id = "house2arm"
        name = "Watch tower armed"

    class Road_outpost(unittype.VehicleType):
        id = "outpost_road"
        name = "Road outpost"

    class Outpost(unittype.VehicleType):
        id = "outpost"
        name = "Outpost"

    class Armed_house(unittype.VehicleType):
        id = "houseA_arm"
        name = "Armed house"


class Infantry:

    class Paratrooper_RPG_16(unittype.VehicleType):
        id = "Paratrooper RPG-16"
        name = "Paratrooper RPG-16"

    class Paratrooper_AKS(unittype.VehicleType):
        id = "Paratrooper AKS-74"
        name = "Paratrooper AKS"

    class Infantry_Soldier_Insurgents(unittype.VehicleType):
        id = "Infantry AK Ins"
        name = "Infantry Soldier Insurgents"

    class Soldier_AK(unittype.VehicleType):
        id = "Soldier AK"
        name = "Soldier AK"

    class Infantry_Soldier_Rus(unittype.VehicleType):
        id = "Infantry AK"
        name = "Infantry Soldier Rus"

    class Soldier_M249(unittype.VehicleType):
        id = "Soldier M249"
        name = "Soldier M249"

    class Infantry_M4(unittype.VehicleType):
        id = "Soldier M4"
        name = "Infantry M4"

    class Georgian_soldier_with_M4(unittype.VehicleType):
        id = "Soldier M4 GRG"
        name = "Georgian soldier with M4"

    class Soldier_RPG(unittype.VehicleType):
        id = "Soldier RPG"
        name = "Soldier RPG"


class Unarmed:

    class GPU_APA_5D_on_Ural_4320(unittype.VehicleType):
        id = "Ural-4320 APA-5D"
        name = "GPU APA-5D on Ural-4320"

    class Fuel_Truck_ATMZ_5(unittype.VehicleType):
        id = "ATMZ-5"
        name = "Fuel Truck ATMZ-5"

    class Fuel_Truck_ATZ_10(unittype.VehicleType):
        id = "ATZ-10"
        name = "Fuel Truck ATZ-10"

    class Transport_GAZ_3307(unittype.VehicleType):
        id = "GAZ-3307"
        name = "Transport GAZ-3307"

    class Transport_GAZ_3308(unittype.VehicleType):
        id = "GAZ-3308"
        name = "Transport GAZ-3308"

    class Transport_GAZ_66(unittype.VehicleType):
        id = "GAZ-66"
        name = "Transport GAZ-66"

    class Tanker_M978_HEMTT(unittype.VehicleType):
        id = "M978 HEMTT Tanker"
        name = "Tanker M978 HEMTT"

    class HEMTT_TFFT(unittype.VehicleType):
        id = "HEMTT TFFT"
        name = "HEMTT TFFT"

    class Transport_IKARUS_280(unittype.VehicleType):
        id = "IKARUS Bus"
        name = "Transport IKARUS-280"

    class Transport_KAMAZ_43101(unittype.VehicleType):
        id = "KAMAZ Truck"
        name = "Transport KAMAZ-43101"

    class Transport_LAZ_695(unittype.VehicleType):
        id = "LAZ Bus"
        name = "Transport LAZ-695"

    class APC_M1025_HMMWV(unittype.VehicleType):
        id = "Hummer"
        name = "APC M1025 HMMWV"
        eprls = True

    class Transport_M818(unittype.VehicleType):
        id = "M 818"
        name = "Transport M818"

    class Transport_MAZ_6303(unittype.VehicleType):
        id = "MAZ-6303"
        name = "Transport MAZ-6303"

    class CP_Predator_GCS(unittype.VehicleType):
        id = "Predator GCS"
        name = "CP Predator GCS"

    class CP_Predator_TrojanSpirit(unittype.VehicleType):
        id = "Predator TrojanSpirit"
        name = "CP Predator TrojanSpirit"

    class Suidae(unittype.VehicleType):
        id = "Suidae"
        name = "Suidae"

    class APC_Tigr_233036(unittype.VehicleType):
        id = "Tigr_233036"
        name = "APC Tigr 233036"

    class Transport_ZIU_9(unittype.VehicleType):
        id = "Trolley bus"
        name = "Transport ZIU-9"

    class Transport_UAZ_469(unittype.VehicleType):
        id = "UAZ-469"
        name = "Transport UAZ-469"

    class Transport_fire_Engine_Ural_ATsP_6(unittype.VehicleType):
        id = "Ural ATsP-6"
        name = "Transport fire-Engine Ural ATsP-6"

    class Transport_Ural_4320_31_Armored(unittype.VehicleType):
        id = "Ural-4320-31"
        name = "Transport Ural-4320-31 Armored"

    class Transport_Ural_4320T(unittype.VehicleType):
        id = "Ural-4320T"
        name = "Transport Ural-4320T"

    class CP_Ural_375_PBU(unittype.VehicleType):
        id = "Ural-375 PBU"
        name = "CP Ural-375 PBU"

    class Transport_Ural_375(unittype.VehicleType):
        id = "Ural-375"
        name = "Transport Ural-375"

    class Transport_VAZ_2109(unittype.VehicleType):
        id = "VAZ Car"
        name = "Transport VAZ-2109"

    class GPU_APA_80_on_ZiL_131(unittype.VehicleType):
        id = "ZiL-131 APA-80"
        name = "GPU APA-80 on ZiL-131"

    class CP_SKP_11_ATC_Mobile_Command_Post(unittype.VehicleType):
        id = "SKP-11"
        name = "CP SKP-11 ATC Mobile Command Post"

    class Transport_ZIL_131_KUNG(unittype.VehicleType):
        id = "ZIL-131 KUNG"
        name = "Transport ZIL-131 KUNG"

    class Transport_ZIL_4331(unittype.VehicleType):
        id = "ZIL-4331"
        name = "Transport ZIL-4331"

    class Transport_KrAZ_6322(unittype.VehicleType):
        id = "KrAZ6322"
        name = "Transport KrAZ-6322"


class Armor:

    class APC_AAV_7(unittype.VehicleType):
        id = "AAV7"
        name = "APC AAV-7"

    class IFV_BMD_1(unittype.VehicleType):
        id = "BMD-1"
        name = "IFV BMD-1"

    class IFV_BMP_1(unittype.VehicleType):
        id = "BMP-1"
        name = "IFV BMP-1"

    class IFV_BMP_2(unittype.VehicleType):
        id = "BMP-2"
        name = "IFV BMP-2"

    class IFV_BMP_3(unittype.VehicleType):
        id = "BMP-3"
        name = "IFV BMP-3"

    class ARV_BRDM_2(unittype.VehicleType):
        id = "BRDM-2"
        name = "ARV BRDM-2"

    class APC_BTR_80(unittype.VehicleType):
        id = "BTR-80"
        name = "APC BTR-80"

    class ARV_BTR_RD(unittype.VehicleType):
        id = "BTR_D"
        name = "ARV BTR-RD"

    class APC_Cobra(unittype.VehicleType):
        id = "Cobra"
        name = "APC Cobra"

    class IFV_LAV_25(unittype.VehicleType):
        id = "LAV-25"
        name = "IFV LAV-25"

    class APC_M1043_HMMWV_Armament(unittype.VehicleType):
        id = "M1043 HMMWV Armament"
        name = "APC M1043 HMMWV Armament"
        eprls = True

    class ATGM_M1045_HMMWV_TOW(unittype.VehicleType):
        id = "M1045 HMMWV TOW"
        name = "ATGM M1045 HMMWV TOW"
        eprls = True

    class APC_M1126_Stryker_ICV(unittype.VehicleType):
        id = "M1126 Stryker ICV"
        name = "APC M1126 Stryker ICV"
        eprls = True

    class APC_M113(unittype.VehicleType):
        id = "M-113"
        name = "APC M113"
        eprls = True

    class ATGM_M1134_Stryker(unittype.VehicleType):
        id = "M1134 Stryker ATGM"
        name = "ATGM M1134 Stryker"
        eprls = True

    class IFV_M2A2_Bradley(unittype.VehicleType):
        id = "M-2 Bradley"
        name = "IFV M2A2 Bradley"
        eprls = True

    class IFV_MCV_80(unittype.VehicleType):
        id = "MCV-80"
        name = "IFV MCV-80"

    class APC_MTLB(unittype.VehicleType):
        id = "MTLB"
        name = "APC MTLB"

    class IFV_Marder(unittype.VehicleType):
        id = "Marder"
        name = "IFV Marder"

    class TPz_Fuchs(unittype.VehicleType):
        id = "TPZ"
        name = "TPz Fuchs"

    class ARV_MTLB_U_BOMAN(unittype.VehicleType):
        id = "Boman"
        name = "ARV MTLB-U BOMAN"

    class MBT_Challenger_II(unittype.VehicleType):
        id = "Challenger2"
        name = "MBT Challenger II"

    class MBT_Leclerc(unittype.VehicleType):
        id = "Leclerc"
        name = "MBT Leclerc"

    class MBT_Leopard_2(unittype.VehicleType):
        id = "Leopard-2"
        name = "MBT Leopard-2"

    class MBT_M60A3_Patton(unittype.VehicleType):
        id = "M-60"
        name = "MBT M60A3 Patton"

    class SPG_M1128_Stryker_MGS(unittype.VehicleType):
        id = "M1128 Stryker MGS"
        name = "SPG M1128 Stryker MGS"
        eprls = True

    class MBT_M1A2_Abrams(unittype.VehicleType):
        id = "M-1 Abrams"
        name = "MBT M1A2 Abrams"
        eprls = True

    class MBT_T_55(unittype.VehicleType):
        id = "T-55"
        name = "MBT T-55"

    class MBT_T_72B(unittype.VehicleType):
        id = "T-72B"
        name = "MBT T-72B"

    class MBT_T_80U(unittype.VehicleType):
        id = "T-80UD"
        name = "MBT T-80U"

    class MBT_T_90(unittype.VehicleType):
        id = "T-90"
        name = "MBT T-90"

    class MBT_Leopard_1A3(unittype.VehicleType):
        id = "Leopard1A3"
        name = "MBT Leopard 1A3"

    class MBT_Merkava_Mk__4(unittype.VehicleType):
        id = "Merkava_Mk4"
        name = "MBT Merkava Mk. 4"


class Artillery:

    class _2B11_mortar(unittype.VehicleType):
        id = "2B11 mortar"
        name = "2B11 mortar"

    class SPH_2S1_Gvozdika(unittype.VehicleType):
        id = "SAU Gvozdika"
        name = "SPH 2S1 Gvozdika"

    class SPH_2S19_Msta(unittype.VehicleType):
        id = "SAU Msta"
        name = "SPH 2S19 Msta"

    class SPH_2S3_Akatsia(unittype.VehicleType):
        id = "SAU Akatsia"
        name = "SPH 2S3 Akatsia"

    class SPH_2S9_Nona(unittype.VehicleType):
        id = "SAU 2-C9"
        name = "SPH 2S9 Nona"

    class SPH_M109_Paladin(unittype.VehicleType):
        id = "M-109"
        name = "SPH M109 Paladin"
        eprls = True

    class MLRS_BM_21_Grad(unittype.VehicleType):
        id = "Grad-URAL"
        name = "MLRS BM-21 Grad"

    class MLRS_9K57_Uragan_BM_27(unittype.VehicleType):
        id = "Uragan_BM-27"
        name = "MLRS 9K57 Uragan BM-27"

    class MLRS_9A52_Smerch(unittype.VehicleType):
        id = "Smerch"
        name = "MLRS 9A52 Smerch"

    class MLRS_M270(unittype.VehicleType):
        id = "MLRS"
        name = "MLRS M270"
        eprls = True


class AirDefence:

    class SAM_SA_19_Tunguska_2S6(unittype.VehicleType):
        id = "2S6 Tunguska"
        name = "SAM SA-19 Tunguska 2S6"

    class SAM_SA_6_Kub_LN_2P25(unittype.VehicleType):
        id = "Kub 2P25 ln"
        name = "SAM SA-6 Kub LN 2P25"

    class SAM_SA_3_S_125_LN_5P73(unittype.VehicleType):
        id = "5p73 s-125 ln"
        name = "SAM SA-3 S-125 LN 5P73"

    class SAM_SA_10_S_300PS_LN_5P85C(unittype.VehicleType):
        id = "S-300PS 5P85C ln"
        name = "SAM SA-10 S-300PS LN 5P85C"

    class SAM_SA_10_S_300PS_LN_5P85D(unittype.VehicleType):
        id = "S-300PS 5P85D ln"
        name = "SAM SA-10 S-300PS LN 5P85D"

    class SAM_SA_11_Buk_LN_9A310M1(unittype.VehicleType):
        id = "SA-11 Buk LN 9A310M1"
        name = "SAM SA-11 Buk LN 9A310M1"

    class SAM_SA_8_Osa_9A33(unittype.VehicleType):
        id = "Osa 9A33 ln"
        name = "SAM SA-8 Osa 9A33"

    class SAM_SA_15_Tor_9A331(unittype.VehicleType):
        id = "Tor 9A331"
        name = "SAM SA-15 Tor 9A331"

    class SAM_SA_13_Strela_10M3_9A35M3(unittype.VehicleType):
        id = "Strela-10M3"
        name = "SAM SA-13 Strela-10M3 9A35M3"

    class SAM_SA_9_Strela_1_9P31(unittype.VehicleType):
        id = "Strela-1 9P31"
        name = "SAM SA-9 Strela-1 9P31"

    class SAM_SA_11_Buk_CC_9S470M1(unittype.VehicleType):
        id = "SA-11 Buk CC 9S470M1"
        name = "SAM SA-11 Buk CC 9S470M1"

    class SAM_SA_8_Osa_LD_9T217(unittype.VehicleType):
        id = "SA-8 Osa LD 9T217"
        name = "SAM SA-8 Osa LD 9T217"

    class SAM_Patriot_AMG_AN_MRC_137(unittype.VehicleType):
        id = "Patriot AMG"
        name = "SAM Patriot AMG AN/MRC-137"

    class SAM_Patriot_ECS_AN_MSQ_104(unittype.VehicleType):
        id = "Patriot ECS"
        name = "SAM Patriot ECS AN/MSQ-104"
        eprls = True

    class SPAAA_Gepard(unittype.VehicleType):
        id = "Gepard"
        name = "SPAAA Gepard"

    class SAM_Hawk_PCP(unittype.VehicleType):
        id = "Hawk pcp"
        name = "SAM Hawk PCP"

    class AAA_Vulcan_M163(unittype.VehicleType):
        id = "Vulcan"
        name = "AAA Vulcan M163"
        eprls = True

    class SAM_Hawk_LN_M192(unittype.VehicleType):
        id = "Hawk ln"
        name = "SAM Hawk LN M192"

    class SAM_Chaparral_M48(unittype.VehicleType):
        id = "M48 Chaparral"
        name = "SAM Chaparral M48"
        eprls = True

    class SAM_Linebacker_M6(unittype.VehicleType):
        id = "M6 Linebacker"
        name = "SAM Linebacker M6"
        eprls = True

    class SAM_Patriot_LN_M901(unittype.VehicleType):
        id = "Patriot ln"
        name = "SAM Patriot LN M901"

    class SAM_Avenger_M1097(unittype.VehicleType):
        id = "M1097 Avenger"
        name = "SAM Avenger M1097"
        eprls = True

    class SAM_Patriot_EPP_III(unittype.VehicleType):
        id = "Patriot EPP"
        name = "SAM Patriot EPP-III"

    class SAM_Patriot_ICC(unittype.VehicleType):
        id = "Patriot cp"
        name = "SAM Patriot ICC"

    class SAM_Roland_ADS(unittype.VehicleType):
        id = "Roland ADS"
        name = "SAM Roland ADS"

    class SAM_SA_10_S_300PS_CP_54K6(unittype.VehicleType):
        id = "S-300PS 54K6 cp"
        name = "SAM SA-10 S-300PS CP 54K6"

    class Stinger_MANPADS(unittype.VehicleType):
        id = "Soldier stinger"
        name = "Stinger MANPADS"

    class SAM_Stinger_comm_dsr(unittype.VehicleType):
        id = "Stinger comm dsr"
        name = "SAM Stinger comm dsr"

    class SAM_Stinger_comm(unittype.VehicleType):
        id = "Stinger comm"
        name = "SAM Stinger comm"

    class SPAAA_ZSU_23_4_Shilka(unittype.VehicleType):
        id = "ZSU-23-4 Shilka"
        name = "SPAAA ZSU-23-4 Shilka"

    class AAA_ZU_23_Closed(unittype.VehicleType):
        id = "ZU-23 Emplacement Closed"
        name = "AAA ZU-23 Closed"

    class AAA_ZU_23_Emplacement(unittype.VehicleType):
        id = "ZU-23 Emplacement"
        name = "AAA ZU-23 Emplacement"

    class AAA_ZU_23_on_Ural_375(unittype.VehicleType):
        id = "Ural-375 ZU-23"
        name = "AAA ZU-23 on Ural-375"

    class AAA_ZU_23_Insurgent_Closed(unittype.VehicleType):
        id = "ZU-23 Closed Insurgent"
        name = "AAA ZU-23 Insurgent Closed"

    class AAA_ZU_23_Insurgent_on_Ural_375(unittype.VehicleType):
        id = "Ural-375 ZU-23 Insurgent"
        name = "AAA ZU-23 Insurgent on Ural-375"

    class AAA_ZU_23_Insurgent(unittype.VehicleType):
        id = "ZU-23 Insurgent"
        name = "AAA ZU-23 Insurgent"

    class SAM_SA_18_Igla_MANPADS(unittype.VehicleType):
        id = "SA-18 Igla manpad"
        name = "SAM SA-18 Igla MANPADS"

    class SAM_SA_18_Igla_comm(unittype.VehicleType):
        id = "SA-18 Igla comm"
        name = "SAM SA-18 Igla comm"

    class SAM_SA_18_Igla_S_MANPADS(unittype.VehicleType):
        id = "SA-18 Igla-S manpad"
        name = "SAM SA-18 Igla-S MANPADS"

    class SAM_SA_18_Igla_S_comm(unittype.VehicleType):
        id = "SA-18 Igla-S comm"
        name = "SAM SA-18 Igla-S comm"

    class SAM_SA_18_Igla_MANPADS(unittype.VehicleType):
        id = "Igla manpad INS"
        name = "SAM SA-18 Igla MANPADS"

    class EWR_1L13(unittype.VehicleType):
        id = "1L13 EWR"
        name = "EWR 1L13"

    class SAM_SA_6_Kub_STR_9S91(unittype.VehicleType):
        id = "Kub 1S91 str"
        name = "SAM SA-6 Kub STR 9S91"

    class SAM_SA_10_S_300PS_TR_30N6(unittype.VehicleType):
        id = "S-300PS 40B6M tr"
        name = "SAM SA-10 S-300PS TR 30N6"

    class SAM_SA_10_S_300PS_SR_5N66M(unittype.VehicleType):
        id = "S-300PS 40B6MD sr"
        name = "SAM SA-10 S-300PS SR 5N66M"

    class EWR_55G6(unittype.VehicleType):
        id = "55G6 EWR"
        name = "EWR 55G6"

    class SAM_SA_10_S_300PS_SR_64H6E(unittype.VehicleType):
        id = "S-300PS 64H6E sr"
        name = "SAM SA-10 S-300PS SR 64H6E"

    class SAM_SA_11_Buk_SR_9S18M1(unittype.VehicleType):
        id = "SA-11 Buk SR 9S18M1"
        name = "SAM SA-11 Buk SR 9S18M1"

    class CP_9S80M1_Sborka(unittype.VehicleType):
        id = "Dog Ear radar"
        name = "CP 9S80M1 Sborka"

    class SAM_Hawk_TR_AN_MPQ_46(unittype.VehicleType):
        id = "Hawk tr"
        name = "SAM Hawk TR AN/MPQ-46"

    class SAM_Hawk_SR_AN_MPQ_50(unittype.VehicleType):
        id = "Hawk sr"
        name = "SAM Hawk SR AN/MPQ-50"
        eprls = True

    class SAM_Patriot_STR_AN_MPQ_53(unittype.VehicleType):
        id = "Patriot str"
        name = "SAM Patriot STR AN/MPQ-53"

    class SAM_Hawk_CWAR_AN_MPQ_55(unittype.VehicleType):
        id = "Hawk cwar"
        name = "SAM Hawk CWAR AN/MPQ-55"
        eprls = True

    class SAM_SA_3_S_125_SR_P_19(unittype.VehicleType):
        id = "p-19 s-125 sr"
        name = "SAM SA-3 S-125 SR P-19"

    class SAM_Roland_EWR(unittype.VehicleType):
        id = "Roland Radar"
        name = "SAM Roland EWR"

    class SAM_SA_3_S_125_TR_SNR(unittype.VehicleType):
        id = "snr s-125 tr"
        name = "SAM SA-3 S-125 TR SNR"

vehicle_map = {
    "2B11 mortar": Artillery._2B11_mortar,
    "SAU Gvozdika": Artillery.SPH_2S1_Gvozdika,
    "SAU Msta": Artillery.SPH_2S19_Msta,
    "SAU Akatsia": Artillery.SPH_2S3_Akatsia,
    "SAU 2-C9": Artillery.SPH_2S9_Nona,
    "M-109": Artillery.SPH_M109_Paladin,
    "AAV7": Armor.APC_AAV_7,
    "BMD-1": Armor.IFV_BMD_1,
    "BMP-1": Armor.IFV_BMP_1,
    "BMP-2": Armor.IFV_BMP_2,
    "BMP-3": Armor.IFV_BMP_3,
    "BRDM-2": Armor.ARV_BRDM_2,
    "BTR-80": Armor.APC_BTR_80,
    "BTR_D": Armor.ARV_BTR_RD,
    "Cobra": Armor.APC_Cobra,
    "LAV-25": Armor.IFV_LAV_25,
    "M1043 HMMWV Armament": Armor.APC_M1043_HMMWV_Armament,
    "M1045 HMMWV TOW": Armor.ATGM_M1045_HMMWV_TOW,
    "M1126 Stryker ICV": Armor.APC_M1126_Stryker_ICV,
    "M-113": Armor.APC_M113,
    "M1134 Stryker ATGM": Armor.ATGM_M1134_Stryker,
    "M-2 Bradley": Armor.IFV_M2A2_Bradley,
    "MCV-80": Armor.IFV_MCV_80,
    "MTLB": Armor.APC_MTLB,
    "Marder": Armor.IFV_Marder,
    "TPZ": Armor.TPz_Fuchs,
    "Boman": Armor.ARV_MTLB_U_BOMAN,
    "Bunker": Fortification.Bunker_2,
    "Paratrooper RPG-16": Infantry.Paratrooper_RPG_16,
    "Paratrooper AKS-74": Infantry.Paratrooper_AKS,
    "Infantry AK Ins": Infantry.Infantry_Soldier_Insurgents,
    "Sandbox": Fortification.Bunker_1,
    "Soldier AK": Infantry.Soldier_AK,
    "Infantry AK": Infantry.Infantry_Soldier_Rus,
    "Soldier M249": Infantry.Soldier_M249,
    "Soldier M4": Infantry.Infantry_M4,
    "Soldier M4 GRG": Infantry.Georgian_soldier_with_M4,
    "Soldier RPG": Infantry.Soldier_RPG,
    "Grad-URAL": Artillery.MLRS_BM_21_Grad,
    "Uragan_BM-27": Artillery.MLRS_9K57_Uragan_BM_27,
    "Smerch": Artillery.MLRS_9A52_Smerch,
    "MLRS": Artillery.MLRS_M270,
    "2S6 Tunguska": AirDefence.SAM_SA_19_Tunguska_2S6,
    "Kub 2P25 ln": AirDefence.SAM_SA_6_Kub_LN_2P25,
    "5p73 s-125 ln": AirDefence.SAM_SA_3_S_125_LN_5P73,
    "S-300PS 5P85C ln": AirDefence.SAM_SA_10_S_300PS_LN_5P85C,
    "S-300PS 5P85D ln": AirDefence.SAM_SA_10_S_300PS_LN_5P85D,
    "SA-11 Buk LN 9A310M1": AirDefence.SAM_SA_11_Buk_LN_9A310M1,
    "Osa 9A33 ln": AirDefence.SAM_SA_8_Osa_9A33,
    "Tor 9A331": AirDefence.SAM_SA_15_Tor_9A331,
    "Strela-10M3": AirDefence.SAM_SA_13_Strela_10M3_9A35M3,
    "Strela-1 9P31": AirDefence.SAM_SA_9_Strela_1_9P31,
    "SA-11 Buk CC 9S470M1": AirDefence.SAM_SA_11_Buk_CC_9S470M1,
    "SA-8 Osa LD 9T217": AirDefence.SAM_SA_8_Osa_LD_9T217,
    "Patriot AMG": AirDefence.SAM_Patriot_AMG_AN_MRC_137,
    "Patriot ECS": AirDefence.SAM_Patriot_ECS_AN_MSQ_104,
    "Gepard": AirDefence.SPAAA_Gepard,
    "Hawk pcp": AirDefence.SAM_Hawk_PCP,
    "Vulcan": AirDefence.AAA_Vulcan_M163,
    "Hawk ln": AirDefence.SAM_Hawk_LN_M192,
    "M48 Chaparral": AirDefence.SAM_Chaparral_M48,
    "M6 Linebacker": AirDefence.SAM_Linebacker_M6,
    "Patriot ln": AirDefence.SAM_Patriot_LN_M901,
    "M1097 Avenger": AirDefence.SAM_Avenger_M1097,
    "Patriot EPP": AirDefence.SAM_Patriot_EPP_III,
    "Patriot cp": AirDefence.SAM_Patriot_ICC,
    "Roland ADS": AirDefence.SAM_Roland_ADS,
    "S-300PS 54K6 cp": AirDefence.SAM_SA_10_S_300PS_CP_54K6,
    "Soldier stinger": AirDefence.Stinger_MANPADS,
    "Stinger comm dsr": AirDefence.SAM_Stinger_comm_dsr,
    "Stinger comm": AirDefence.SAM_Stinger_comm,
    "ZSU-23-4 Shilka": AirDefence.SPAAA_ZSU_23_4_Shilka,
    "ZU-23 Emplacement Closed": AirDefence.AAA_ZU_23_Closed,
    "ZU-23 Emplacement": AirDefence.AAA_ZU_23_Emplacement,
    "Ural-375 ZU-23": AirDefence.AAA_ZU_23_on_Ural_375,
    "ZU-23 Closed Insurgent": AirDefence.AAA_ZU_23_Insurgent_Closed,
    "Ural-375 ZU-23 Insurgent": AirDefence.AAA_ZU_23_Insurgent_on_Ural_375,
    "ZU-23 Insurgent": AirDefence.AAA_ZU_23_Insurgent,
    "SA-18 Igla manpad": AirDefence.SAM_SA_18_Igla_MANPADS,
    "SA-18 Igla comm": AirDefence.SAM_SA_18_Igla_comm,
    "SA-18 Igla-S manpad": AirDefence.SAM_SA_18_Igla_S_MANPADS,
    "SA-18 Igla-S comm": AirDefence.SAM_SA_18_Igla_S_comm,
    "Igla manpad INS": AirDefence.SAM_SA_18_Igla_MANPADS,
    "1L13 EWR": AirDefence.EWR_1L13,
    "Kub 1S91 str": AirDefence.SAM_SA_6_Kub_STR_9S91,
    "S-300PS 40B6M tr": AirDefence.SAM_SA_10_S_300PS_TR_30N6,
    "S-300PS 40B6MD sr": AirDefence.SAM_SA_10_S_300PS_SR_5N66M,
    "55G6 EWR": AirDefence.EWR_55G6,
    "S-300PS 64H6E sr": AirDefence.SAM_SA_10_S_300PS_SR_64H6E,
    "SA-11 Buk SR 9S18M1": AirDefence.SAM_SA_11_Buk_SR_9S18M1,
    "Dog Ear radar": AirDefence.CP_9S80M1_Sborka,
    "Hawk tr": AirDefence.SAM_Hawk_TR_AN_MPQ_46,
    "Hawk sr": AirDefence.SAM_Hawk_SR_AN_MPQ_50,
    "Patriot str": AirDefence.SAM_Patriot_STR_AN_MPQ_53,
    "Hawk cwar": AirDefence.SAM_Hawk_CWAR_AN_MPQ_55,
    "p-19 s-125 sr": AirDefence.SAM_SA_3_S_125_SR_P_19,
    "Roland Radar": AirDefence.SAM_Roland_EWR,
    "snr s-125 tr": AirDefence.SAM_SA_3_S_125_TR_SNR,
    "house1arm": Fortification.Barracks_armed,
    "house2arm": Fortification.Watch_tower_armed,
    "outpost_road": Fortification.Road_outpost,
    "outpost": Fortification.Outpost,
    "houseA_arm": Fortification.Armed_house,
    "Challenger2": Armor.MBT_Challenger_II,
    "Leclerc": Armor.MBT_Leclerc,
    "Leopard-2": Armor.MBT_Leopard_2,
    "M-60": Armor.MBT_M60A3_Patton,
    "M1128 Stryker MGS": Armor.SPG_M1128_Stryker_MGS,
    "M-1 Abrams": Armor.MBT_M1A2_Abrams,
    "T-55": Armor.MBT_T_55,
    "T-72B": Armor.MBT_T_72B,
    "T-80UD": Armor.MBT_T_80U,
    "T-90": Armor.MBT_T_90,
    "Leopard1A3": Armor.MBT_Leopard_1A3,
    "Merkava_Mk4": Armor.MBT_Merkava_Mk__4,
    "Ural-4320 APA-5D": Unarmed.GPU_APA_5D_on_Ural_4320,
    "ATMZ-5": Unarmed.Fuel_Truck_ATMZ_5,
    "ATZ-10": Unarmed.Fuel_Truck_ATZ_10,
    "GAZ-3307": Unarmed.Transport_GAZ_3307,
    "GAZ-3308": Unarmed.Transport_GAZ_3308,
    "GAZ-66": Unarmed.Transport_GAZ_66,
    "M978 HEMTT Tanker": Unarmed.Tanker_M978_HEMTT,
    "HEMTT TFFT": Unarmed.HEMTT_TFFT,
    "IKARUS Bus": Unarmed.Transport_IKARUS_280,
    "KAMAZ Truck": Unarmed.Transport_KAMAZ_43101,
    "LAZ Bus": Unarmed.Transport_LAZ_695,
    "Hummer": Unarmed.APC_M1025_HMMWV,
    "M 818": Unarmed.Transport_M818,
    "MAZ-6303": Unarmed.Transport_MAZ_6303,
    "Predator GCS": Unarmed.CP_Predator_GCS,
    "Predator TrojanSpirit": Unarmed.CP_Predator_TrojanSpirit,
    "Suidae": Unarmed.Suidae,
    "Tigr_233036": Unarmed.APC_Tigr_233036,
    "Trolley bus": Unarmed.Transport_ZIU_9,
    "UAZ-469": Unarmed.Transport_UAZ_469,
    "Ural ATsP-6": Unarmed.Transport_fire_Engine_Ural_ATsP_6,
    "Ural-4320-31": Unarmed.Transport_Ural_4320_31_Armored,
    "Ural-4320T": Unarmed.Transport_Ural_4320T,
    "Ural-375 PBU": Unarmed.CP_Ural_375_PBU,
    "Ural-375": Unarmed.Transport_Ural_375,
    "VAZ Car": Unarmed.Transport_VAZ_2109,
    "ZiL-131 APA-80": Unarmed.GPU_APA_80_on_ZiL_131,
    "SKP-11": Unarmed.CP_SKP_11_ATC_Mobile_Command_Post,
    "ZIL-131 KUNG": Unarmed.Transport_ZIL_131_KUNG,
    "ZIL-4331": Unarmed.Transport_ZIL_4331,
    "KrAZ6322": Unarmed.Transport_KrAZ_6322,
}



from .unitgroup import VehicleGroup, ShipGroup, PlaneGroup, StaticGroup, HelicopterGroup, FlyingGroup
from typing import List, Dict


def find_exact(group_name, find_name):
    return group_name == find_name


def find_match(group_name, find_name):
    return find_name in group_name

find_map = {
    "exact": find_exact,
    "match": find_match
}


class Country:
    callsign = {}
    planes = []
    helicopters = []

    def __init__(self, _id, name):
        self.id = _id
        self.name = name
        self.vehicle_group = []  # type: List[VehicleGroup]
        self.ship_group = []  # type: List[ShipGroup]
        self.plane_group = []  # type: List[PlaneGroup]
        self.helicopter_group = []  # type: List[HelicopterGroup]
        self.static_group = []  # type: List[StaticGroup]
        self.current_callsign_id = 99
        self.current_callsign_category = {}  # type: Dict[str,int]

    def add_vehicle_group(self, vgroup):
        self.vehicle_group.append(vgroup)

    def add_ship_group(self, sgroup):
        self.ship_group.append(sgroup)

    def add_plane_group(self, pgroup):
        self.plane_group.append(pgroup)

    def add_helicopter_group(self, hgroup):
        self.helicopter_group.append(hgroup)

    def add_aircraft_group(self, group: FlyingGroup):
        if group.units[0].unit_type.helicopter:
            self.helicopter_group.append(group)
        else:
            self.plane_group.append(group)

    def add_static_group(self, sgroup):
        self.static_group.append(sgroup)

    def find_group(self, group_name, search="exact"):
        groups = [self.vehicle_group,
                  self.ship_group,
                  self.plane_group,
                  self.helicopter_group,
                  self.static_group]
        for search_group in groups:
            for group in search_group:
                if find_map[search](group.name.str(), group_name):
                    return group
        return None

    def find_vehicle_group(self, name: str, search="exact"):
        for group in self.vehicle_group:
            if find_map[search](group.name.str(), name):
                return group
        return None

    def find_ship_group(self, name: str, search="exact"):
        for group in self.ship_group:
            if find_map[search](group.name.str(), name):
                return group
        return None

    def find_plane_group(self, name: str, search="exact"):
        for group in self.plane_group:
            if find_map[search](group.name.str(), name):
                return group

    def find_helicopter_group(self, name: str, search="exact"):
        for group in self.helicopter_group:
            if find_map[search](group.name.str(), name):
                return group

    def find_static_group(self, name: str, search="exact"):
        for group in self.static_group:
            if find_map[search](group.name.str(), name):
                return group
        return None

    def next_callsign_id(self):
        self.current_callsign_id += 1
        return self.current_callsign_id

    def next_callsign_category(self, category):
        if category not in self.current_callsign_category:
            self.current_callsign_category[category] = 0
            return self.callsign.get(category)[0]

        self.current_callsign_category[category] += 1
        if self.current_callsign_category[category] >= len(self.callsign[category]):
            self.current_callsign_category[category] = 0
        return self.callsign.get(category)[self.current_callsign_category[category]]

    def dict(self):
        d = {
            "name": self.name,
            "id": self.id
        }

        if self.vehicle_group:
            d["vehicle"] = {"group": {}}
            i = 1
            for vgroup in self.vehicle_group:
                d["vehicle"]["group"][i] = vgroup.dict()
                i += 1

        if self.ship_group:
            d["ship"] = {"group": {}}
            i = 1
            for group in self.ship_group:
                d["ship"]["group"][i] = group.dict()
                i += 1

        if self.plane_group:
            d["plane"] = {"group": {}}
            i = 1
            for plane_group in self.plane_group:
                d["plane"]["group"][i] = plane_group.dict()
                i += 1

        if self.helicopter_group:
            d["helicopter"] = {"group": {}}
            i = 1
            for group in self.helicopter_group:
                d["helicopter"]["group"][i] = group.dict()
                i += 1

        if self.static_group:
            d["static"] = {"group": {}}
            i = 1
            for static_group in self.static_group:
                d["static"]["group"][i] = static_group.dict()
                i += 1
        return d

    def __str__(self):
        return str(self.id) + "," + self.name + "," + str(self.vehicle_group)



def loads(tablestr):

    class Parser:

        def __init__(self, buffer: str):
            self.buffer = buffer
            self.buflen = len(buffer)
            self.pos = 0  # type: int
            self.lineno = 1  # type: int

        def value(self):
            self.eat_ws()

            c = self.buffer[self.pos]
            if c == '{':
                return self.object()
            elif c == '"':
                return self.string()
            elif c.isnumeric() or c == '-':
                return self.number()
            elif c == '_':
                return self.str_function()
            else:  # varname
                varname = self.eatvarname()
                if varname == 'false' or varname == 'true':
                    return varname == 'true'
                if varname == 'local':
                    # ignore local keyword
                    self.eat_ws()
                    varname = self.eatvarname()
                self.eat_ws()
                if not self.eob() and self.buffer[self.pos] == '=':
                    self.pos += 1
                    return {varname: self.value()}
                else:
                    se = SyntaxError()
                    se.text = varname + " '" + self.buffer[self.pos] + "'"
                    se.lineno = self.lineno
                    se.offset = self.pos
                    raise se

            return {}

        def str_function(self):
            if self.buffer[self.pos] != '_':
                se = SyntaxError()
                se.lineno = self.lineno
                se.offset = self.pos
                se.text = "Expected character '_', got '{char}'".format(char=self.buffer[self.pos])
                raise se

            if self.advance():
                raise self.eob_exception()

            self.eat_ws()

            if self.buffer[self.pos] != '(':
                se = SyntaxError()
                se.lineno = self.lineno
                se.offset = self.pos
                se.text = "Expected character '(', got '{char}'".format(char=self.buffer[self.pos])
                raise se

            self.advance()
            self.eat_ws()

            s = self.string()

            self.eat_ws()

            if self.buffer[self.pos] != ')':
                se = SyntaxError()
                se.lineno = self.lineno
                se.offset = self.pos
                se.text = "Expected character ')', got '{char}'".format(char=self.buffer[self.pos])
                raise se

            self.pos += 1
            return s


        def string(self):
            if self.buffer[self.pos] != '"':
                se = SyntaxError()
                se.lineno = self.lineno
                se.offset = self.pos
                se.text = "Expected character '\"', got '{char}'".format(char=self.buffer[self.pos])
                raise se

            state = 0
            s = ''
            while state != 1:
                if self.advance():
                    raise self.eob_exception()

                c = self.buffer[self.pos]
                if state == 0:
                    if c == '"':
                        state = 1
                        self.pos += 1
                    elif c == '\\':
                        state = 2
                    else:
                        s += c
                elif state == 2:
                    s += c
                    state = 0
            return s

        def number(self):
            n = ''
            sign = 1
            if self.buffer[self.pos] == '-':
                sign = -1
                if self.advance():
                    raise self.eob_exception()

            while (not self.eob() and
                    (self.buffer[self.pos].isnumeric() or self.buffer[self.pos] == '.' or
                        self.buffer[self.pos].lower() == 'e')):
                n += self.buffer[self.pos]
                self.pos += 1

            num = float(n) * sign
            if num.is_integer():
                return int(num)
            return num

        def object(self):
            d = {}
            if self.buffer[self.pos] != '{':
                se = SyntaxError()
                se.lineno = self.lineno
                se.offset = self.pos
                se.text = "Expected character '{', got '{char}'".format(char=self.buffer[self.pos])
                raise se

            if self.advance():
                raise self.eob_exception()

            self.eat_ws()

            while self.buffer[self.pos] != '}':
                self.eat_ws()

                if self.buffer[self.pos] != '[':
                    se = SyntaxError()
                    se.lineno = self.lineno
                    se.offset = self.pos
                    se.text = "Expected character '[', got '{char}'".format(char=self.buffer[self.pos])
                    raise se

                if self.advance():
                    raise self.eob_exception()

                self.eat_ws()
                if self.buffer[self.pos] == '"':
                    key = self.string()
                else:
                    key = self.number()

                if self.eob():
                    raise self.eob_exception()

                self.eat_ws()

                if self.buffer[self.pos] != ']':
                    se = SyntaxError()
                    se.lineno = self.lineno
                    se.offset = self.pos
                    se.text = "Expected character ']', got '{char}'".format(char=self.buffer[self.pos])
                    raise se

                if self.advance():
                    raise self.eob_exception()

                self.eat_ws()

                if self.buffer[self.pos] != '=':
                    se = SyntaxError()
                    se.lineno = self.lineno
                    se.offset = self.pos
                    se.text = "Expected character '=', got '{char}'".format(char=self.buffer[self.pos])
                    raise se

                if self.advance():
                    raise self.eob_exception()

                val = self.value()

                self.eat_ws()

                d[key] = val
                # print(key, ':', val)

                c = self.buffer[self.pos]
                if c == '}':
                    break
                elif c == ',':
                    if self.advance():
                        raise self.eob_exception()
                    self.eat_ws()
                else:
                    se = SyntaxError()
                    se.lineno = self.lineno
                    se.offset = self.pos
                    se.text = "Unexpected character '{char}'".format(char=self.buffer[self.pos])
                    raise se

            self.pos += 1

            return d

        def eatvarname(self):
            varname = ''
            while (not self.eob()) and self.buffer[self.pos].isalnum():
                varname += self.buffer[self.pos]
                self.pos += 1

            return varname

        def eat_comment(self):
            if (self.buffer[self.pos] == '-' and
                self.pos + 1 < self.buflen and
                self.buffer[self.pos + 1] == '-'):
                while not self.eob() and self.buffer[self.pos] != '\n':
                    self.pos += 1

        def eat_ws(self):
            self.eat_comment()
            while True:
                if self.pos >= self.buflen:
                    return
                c = self.buffer[self.pos]  # type: str
                if c == '\n':
                    self.lineno += 1
                if c == '-':
                    self.eat_comment()
                    c = self.buffer[self.pos]
                if not c.isspace():
                    return

                self.pos += 1

        def eob(self):
            return self.pos >= self.buflen

        def eob_exception(self):
            se = SyntaxError()
            se.lineno = self.lineno
            se.offset = self.pos
            se.text = "Unexpected end of buffer"
            return se

        def char(self):
            return self.buffer[self.pos]

        def advance(self):
            self.pos += 1
            return self.eob()

    p = Parser(tablestr)
    return p.value()



from . import mission
from . import task
from . import templates
from . import countries
from . import planes
from . import helicopters
from . import statics
from . import vehicles
from . import ships
from . import terrain
from . import unit
from . import unitgroup
from . import goals
from . import weather
from . import point
from . import triggers
from . import condition
from . import action
from . import forcedoptions
from .mapping import Point, Rectangle, Polygon
from .mission import Mission



from datetime import datetime, timezone
import random
import math
from enum import Enum
from typing import List
from . import mapping


class Wind:
    def __init__(self, direction=0, speed=0):
        self.direction = direction
        self.speed = speed

    def dict(self):
        return {"speed": self.speed, "dir": self.direction}


class Cyclone:
    def __init__(self):
        self.pressure_spread = 0.0
        self.centerZ = 0.0
        self.ellipticity = 0.0
        self.rotation = 0.0
        self.pressure_excess = 0
        self.centerX = 0.0

    def dict(self):
        d = {
            "pressure_spread": self.pressure_spread,
            "pressure_excess": self.pressure_excess,
            "centerZ": self.centerZ,
            "ellipticity": self.ellipticity,
            "rotation": self.rotation,
            "centerX": self.centerX
        }
        return d

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self.dict()) + ')'


class Weather:
    class BaricSystem(Enum):
        Cyclone = 0
        AntiCyclone = 1
        None_ = 2

    class Preceptions(Enum):
        None_ = 0
        Rain = 1
        Thunderstorm = 2

    def __init__(self, terrain):
        self.terrain = terrain
        self.atmosphere_type = 0
        self.wind_at_ground = Wind()
        self.wind_at_2000 = Wind()
        self.wind_at_8000 = Wind()
        self.enable_fog = False
        self.turbulence_at_ground = 0
        self.turbulence_at_2000 = 0
        self.turbulence_at_8000 = 0
        self.season_temperature = 20
        self.type_weather = 0
        self.qnh = 760
        self.cyclones = []
        self.name = "Summer, clean sky"
        self.fog_thickness = 0
        self.fog_visibility = 25
        self.fog_density = 7
        self.visibility_distance = 80000
        self.clouds_thickness = 200
        self.clouds_density = 0
        self.clouds_base = 300
        self.clouds_iprecptns = Weather.Preceptions.None_

    def load_from_dict(self, d):
        self.atmosphere_type = d["atmosphere_type"]
        wind = d.get("wind", {})
        wind_at_ground = wind.get("atGround", {})
        wind_at_2000 = wind.get("at2000", {})
        wind_at_8000 = wind.get("at8000", {})
        self.wind_at_ground = Wind(wind_at_ground.get("dir", 0), wind_at_ground.get("speed", 0))
        self.wind_at_2000 = Wind(wind_at_2000.get("dir", 0), wind_at_2000.get("speed", 0))
        self.wind_at_8000 = Wind(wind_at_8000.get("dir", 0), wind_at_8000.get("speed", 0))
        self.enable_fog = d["enable_fog"]
        turbulence = d.get("turbulence", {})
        self.turbulence_at_ground = turbulence.get("atGround", 0)
        self.turbulence_at_2000 = turbulence.get("at2000", 0)
        self.turbulence_at_8000 = turbulence.get("at8000", 0)
        season = d.get("season", {})
        self.season_temperature = season.get("temperature", 20)
        self.type_weather = d.get("type_weather", 0)
        self.qnh = d.get("qnh", 760)
        cyclones = d.get("cyclones", {})
        for x in cyclones:
            c = Cyclone()
            c.centerX = cyclones[x].get("centerX", 0)
            c.centerZ = cyclones[x].get("centerZ", 0)
            c.ellipticity = cyclones[x].get("ellipticity", 0)
            c.pressure_excess = cyclones[x].get("pressure_excess", 0)
            c.pressure_spread = cyclones[x].get("pressure_spread", 0)
            c.rotation = cyclones[x].get("rotation", 0)
            self.cyclones.append(c)
        self.name = d.get("name", "Summer, clean sky")
        fog = d.get("fog", {})
        self.fog_thickness = fog.get("thickness", 0)
        self.fog_visibility = fog.get("visibility", 25)
        self.fog_density = fog.get("density", 7)
        visibility = d.get("visiblity", {})
        self.visibility_distance = visibility.get("distance", 80000)
        clouds = d.get("clouds", {})
        self.clouds_thickness = clouds.get("thickness", 200)
        self.clouds_density = clouds.get("density", 0)
        self.clouds_base = clouds.get("base", 300)
        self.clouds_iprecptns = Weather.Preceptions(clouds.get("iprecptns", 0))

    def set_season_from_datetime(self, dt: datetime):
        if datetime(dt.year, 3, 1, tzinfo=timezone.utc) <= dt < datetime(dt.year, 6, 1, tzinfo=timezone.utc):
            self.season_temperature = random.randrange(11, 22)
        elif datetime(dt.year, 6, 1, tzinfo=timezone.utc) <= dt < datetime(dt.year, 9, 1, tzinfo=timezone.utc):
            self.season_temperature = random.randrange(18, 35)
        elif datetime(dt.year, 9, 1, tzinfo=timezone.utc) <= dt < datetime(dt.year, 12, 1, tzinfo=timezone.utc):
            self.season_temperature = random.randrange(10, 20)
        else:
            self.season_temperature = random.randrange(-6, 6)

    @staticmethod
    def random_normals() -> List[float]:
        rands = []
        for x in range(0, 3):
            r = max(0.0000001, random.random())
            fi = max(0.0000001, random.random())

            rands.append(math.cos(math.pi * 2 * fi) * math.sqrt(-2 * math.log(r)))
            rands.append(math.sin(math.pi * 2 * fi) * math.sqrt(-2 * math.log(r)))

        return rands

    @staticmethod
    def _init_cyclone(rands, idx, params, wtype, position):
        mainpeakexcessmod = 1

        if wtype == Weather.BaricSystem.None_ or (wtype != Weather.BaricSystem.None_ and idx != 1):
            x = math.cos(params['initangle'] + params['anglestep'] * (idx - 1) + params['deltaangle'] * rands[0])
            x *= params['distance'] + params['distancestddev'] * rands[1]
            y = math.sin(params['initangle'] + params['anglestep'] * (idx - 1) + params['deltaangle'] * rands[0])
            y *= params['distance'] + params['distancestddev'] * rands[1]
        else:
            mainpeakexcessmod = 0.22
            x = math.cos(params['initangle'] + params['anglestep'] * (idx - 1) + params['deltaangle'] * rands[0])
            x *= params['distancestddev'] * mainpeakexcessmod * rands[1]
            y = math.sin(params['initangle'] + params['anglestep'] * (idx - 1) + params['deltaangle'] * rands[0])
            y *= params['distancestddev'] * mainpeakexcessmod * rands[1]

        c = Cyclone()
        c.centerX = position.x + x
        c.centerZ = position.y + y
        c.ellipticity = 1 + rands[2] * params['ellipticitystddev']
        c.pressure_excess = math.floor(params['sign'] * abs(rands[3]) *
                                       params['pressurestddev'] * mainpeakexcessmod + params['pressureoffset'])
        c.rotation = params['rotationstddev'] * rands[4]
        c.pressure_spread = params['spread'] + params['spreadstddev'] * mainpeakexcessmod * rands[5]

        if wtype == Weather.BaricSystem.AntiCyclone:
            c.pressure_spread *= 1.3
        return c

    def dynamic_weather(self, system: BaricSystem, cyclones: int=1):
        self.cyclones.clear()

        center = self.terrain.bounds.center()  # type: mapping.Point
        self.atmosphere_type = 1
        self.type_weather = system.value

        params = {
            'initangle': 2 * math.pi * random.random(),
            'distance': 950000,
            'distancestddev': 150000,
            'spread': 900000,
            'spreadstddev': 150000,
            'pressureoffset': 1200,
            'pressurestddev': 500,
            'ellipticitystddev': 0.25,
            'rotationstddev': 1.0471975511965977461542144610932
        }

        if system == Weather.BaricSystem.None_:
            params['anglestep'] = 2 * math.pi * cyclones
            params['deltaangle'] = params['anglestep'] / 4

            for i in range(1, cyclones + 1):
                params['sign'] = random.randrange(0, 1) * 2 - 1
                self.cyclones.append(Weather._init_cyclone(Weather.random_normals(), i, params, system, center))

        else:
            params['anglestep'] = 2 * math.pi / (cyclones - 1) if cyclones > 1 else 0
            params['deltaangle'] = params['anglestep'] / 4

            params['sign'] = -1 if system == Weather.BaricSystem.Cyclone else 1

            for i in range(1, cyclones + 1):
                self.cyclones.append(Weather._init_cyclone(Weather.random_normals(), i, params, system, center))
                params['sign'] = random.randrange(0, 1) * 2 - 1

        min_pressure = min([x.pressure_excess for x in self.cyclones])
        max_pressure = max([x.pressure_excess for x in self.cyclones])
        pressure_diff = max_pressure - min_pressure
        # min_pressure = 0
        # for c in range(0, cyclones):
        #     # TODO ask ED, if we are allowed to use generateCyclones code
        #     c = Cyclone()
        #     pos = center.point_from_heading(random.randrange(0, 360), random.randrange(10*1000, 1000*1000, 1000))
        #     c.centerZ = pos.y
        #     c.centerX = pos.x
        #     c.pressure_spread = random.random() * 100000 + 800000
        #     c.rotation = random.random() * math.pi
        #
        #     c.pressure_excess = random.randrange(1200, 1800) if random.random() > 0.8 else random.randrange(600, 1300)
        #
        #     c.ellipticity = 1 + random.random() * 0.25
        #
        #     if system in [Weather.BaricSystem.Cyclone, Weather.BaricSystem.None_]:
        #         c.pressure_excess *= -1
        #
        #     min_pressure = min(min_pressure, c.pressure_excess)
        #
        #     self.cyclones.append(c)

        if pressure_diff < 200:
            self.turbulence_at_ground = random.randrange(0, 10)
        elif 200 <= min_pressure < 600:
            self.turbulence_at_ground = random.randrange(10, 25)
        else:
            self.turbulence_at_ground = random.randrange(25, 60)

    def _random_thunderstorm(self):
        self.atmosphere_type = 0
        wind_dir = random.randrange(0, 359) + 180
        wind_speed = random.randrange(10, 30)
        self.wind_at_ground.direction = (wind_dir + random.randrange(-90, 90) - 180) % 360
        self.wind_at_ground.speed = wind_speed + random.randrange(-5, 10)
        self.wind_at_2000.direction = (wind_dir + random.randrange(-90, 90) - 180) % 360
        self.wind_at_2000.speed = wind_speed + random.randrange(-5, 10)
        self.wind_at_8000.direction = (wind_dir + random.randrange(-90, 90) - 180) % 360
        self.wind_at_8000.speed = wind_speed + random.randrange(-5, 10)

        self.clouds_iprecptns = Weather.Preceptions.Thunderstorm
        self.clouds_base = random.randrange(800, 1000)
        self.clouds_density = 10
        self.clouds_thickness = random.randrange(300, 900)

        self.turbulence_at_ground = random.randrange(30, 70)
        self.turbulence_at_2000 = random.randrange(2, 10)
        self.turbulence_at_8000 = random.randrange(2, 10)

        self.qnh = random.randrange(715, 750)
        self.visibility_distance = 80000

        fog_base = min(0.8, random.random())
        self.enable_fog = True
        self.fog_density = 3
        self.fog_thickness = int(700 * fog_base)
        self.fog_visibility = int(5500 - (4000 * fog_base))

    def random(self, dt: datetime=None):
        # check if there might be the season for thunderstorms
        if 4 < dt.month < 11:
            if random.random() > 0.9:
                self._random_thunderstorm()
                return

        self.dynamic_weather(random.choice(list(Weather.BaricSystem)), random.randrange(1, 4))

    def dict(self):
        d = {}
        d["atmosphere_type"] = self.atmosphere_type
        d["wind"] = {"atGround": self.wind_at_ground.dict(),
                     "at2000": self.wind_at_2000.dict(),
                     "at8000": self.wind_at_8000.dict()}
        d["enable_fog"] = self.enable_fog
        d["turbulence"] = {"atGround": self.turbulence_at_ground,
                           "at2000": self.turbulence_at_2000,
                           "at8000": self.turbulence_at_8000}
        d["season"] = {"temperature": self.season_temperature}
        d["type_weather"] = self.type_weather
        d["qnh"] = self.qnh
        d["cyclones"] = {x+1: self.cyclones[x].dict() for x in range(0, len(self.cyclones))}
        d["name"] = self.name
        d["fog"] = {"thickness": self.fog_thickness, "visibility": self.fog_visibility, "density": self.fog_density}
        d["visibility"] = {"distance": self.visibility_distance}
        d["clouds"] = {"thickness": self.clouds_thickness,
                       "density": self.clouds_density,
                       "base": self.clouds_base,
                       "iprecptns": self.clouds_iprecptns.value}
        return d



"""
The mission module is the entry point to all pydcs functions.
"""
import copy
import os
import sys
import tempfile
import zipfile
import random
from datetime import datetime, timezone, timedelta
from enum import Enum
from typing import List, Dict, Union, Optional

from dcs.coalition import Coalition
from dcs.terrain.terrain import Warehouses
from dcs.triggers import Triggers
from . import countries
from . import helicopters
from . import lua
from . import mapping
from . import planes
from . import task
from . import unitgroup
from . import unittype
from . import weather
from . import triggers
from .country import Country
from .forcedoptions import ForcedOptions
from .goals import Goals
from .groundcontrol import GroundControl
from .point import StaticPoint, MovingPoint, PointAction, PointProperties
from .terrain import Caucasus, Nevada, ParkingSlot, Airport, NoParkingSlotError
from .translation import Translation
from .unit import Plane, Helicopter, Ship, Vehicle, Static


class StartType(Enum):
    """Enum class for start types."""
    Cold = 1
    """Coldstart from ramp."""
    Warm = 2
    """Warmstart from ramp."""
    Runway = 3
    """Start from runway."""

    @staticmethod
    def from_string(s: str):
        """Returns the StartType enum for a string value.

        ["cold", "warm", "runway"]

        Args:
            s: string representation of the starttype

        Returns:
            the correct StartType.
        """
        st_map = {
            "cold": StartType.Cold,
            "warm": StartType.Warm,
            "runway": StartType.Runway
        }
        return st_map[s.lower()]


class Mission:
    """This class represents the whole dcs .miz file.

    A .miz file is a zip file containing all needed files to run a mission.
    example.miz:

      * mission
      * options
      * warehouses
      * i10n

        * DEFAULT
        * dictionary

          * mapResource
          * [localized resource files, .wav, .jpg, ...]

    Args:
        terrain: the used terrain for this mission.
    """
    _COUNTRY_IDS = {x for x in range(0, 13)} | {x for x in range(15, 47)}

    def __init__(self, terrain: Union[Caucasus, Nevada]=None):
        if terrain is None:
            terrain = Caucasus()

        self.current_unit_id = 0
        self.current_group_id = 0
        self.current_dict_id = 0
        self.filename = None

        self.translation = Translation(self)
        self.map_resource = MapResource(self)

        self._description_text = self.string("")
        self._description_bluetask = self.string("")
        self._description_redtask = self.string("")
        self._sortie = self.string("")
        self.pictureFileNameR = []
        self.pictureFileNameB = []
        self.version = 12
        self.currentKey = 0
        self.start_time = datetime.fromtimestamp(1306886400 + 43200, timezone.utc)  # 01-06-2011 12:00:00 UTC
        self.season_from_start_time = True
        """If set to True the mission season will be set by the value of :py:attr:`Mission.start_time`"""
        self.random_weather = False
        """If set to True a random weather will be generated"""
        self.terrain = terrain
        self.triggerrules = triggers.Rules()
        self.triggers = Triggers()
        self.init_script_file = None
        self.init_script = None
        self.options = Options()
        self.warehouses = Warehouses(self.terrain)
        self.goals = Goals()
        blue = Coalition("blue")
        blue.add_country(countries.Australia())
        blue.add_country(countries.Belgium())
        blue.add_country(countries.Canada())
        blue.add_country(countries.Croatia())
        blue.add_country(countries.CzechRepublic())
        blue.add_country(countries.Denmark())
        blue.add_country(countries.France())
        blue.add_country(countries.Georgia())
        blue.add_country(countries.Germany())
        blue.add_country(countries.Israel())
        blue.add_country(countries.Italy())
        blue.add_country(countries.Norway())
        blue.add_country(countries.Poland())
        blue.add_country(countries.SouthKorea())
        blue.add_country(countries.Spain())
        blue.add_country(countries.TheNetherlands())
        blue.add_country(countries.UK())
        blue.add_country(countries.USA())
        blue.add_country(countries.Turkey())

        red = Coalition("red")
        red.add_country(countries.Abkhazia())
        red.add_country(countries.Belarus())
        red.add_country(countries.China())
        red.add_country(countries.Iran())
        red.add_country(countries.Kazakhstan())
        red.add_country(countries.NorthKorea())
        red.add_country(countries.Russia())
        red.add_country(countries.Serbia())
        red.add_country(countries.SouthOssetia())
        red.add_country(countries.Syria())
        red.add_country(countries.Ukraine())

        blue.bullseye = terrain.bullseye_blue
        red.bullseye = terrain.bullseye_red

        self.coalition = {"blue": blue, "red": red}  # type: Dict[str, Coalition]

        self.map = self.terrain.map_view_default  # type: terrain.MapView

        self.failures = {}
        self.groundControl = GroundControl()
        self.forced_options = ForcedOptions()
        self.resourceCounter = {}  # keep default or empty, old format
        self.needModules = {}
        self.weather = weather.Weather(self.terrain)
        # TODO used modules
        self.usedModules = {
            'Su-25A by Eagle Dynamics': True,
            'MiG-21Bis AI by Leatherneck Simulations': True,
            'UH-1H Huey by Belsimtek': True,
            'Su-25T by Eagle Dynamics': True,
            'F-86F Sabre by Belsimtek': True,
            'Su-27 Flanker by Eagle Dynamics': True,
            'Hawk T.1A AI by VEAO Simulations': True,
            'MiG-15bis AI by Eagle Dynamics': True,
            'Ka-50 Black Shark by Eagle Dynamics': True,
            'Combined Arms by Eagle Dynamics': True,
            'L-39C/ZA by Eagle Dynamics': True,
            'A-10C Warthog by Eagle Dynamics': True,
            'F-5E/E-3 by Belsimtek': True,
            'C-101 Aviojet': True,
            'TF-51D Mustang by Eagle Dynamics': True,
            './CoreMods/aircraft/MQ-9 Reaper': True,
            'C-101 Aviojet by AvioDev': True,
            'P-51D Mustang by Eagle Dynamics': True,
            'A-10A by Eagle Dynamics': True,
            'L-39C': True,
            'World War II AI Units by Eagle Dynamics': True,
            'MiG-15bis by Belsimtek': True,
            'F-15C': True,
            'Flaming Cliffs by Eagle Dynamics': True,
            'Bf 109 K-4 by Eagle Dynamics': True,
            'Mi-8MTV2 Hip by Belsimtek': True,
            'MiG-21Bis by Leatherneck Simulations': True,
            'M-2000C by RAZBAM Sims': True,
            'M-2000C AI by RAZBAM Sims': True,
            'FW-190D9 Dora by Eagle Dynamics': True,
            'Caucasus': True,
            'Hawk T.1A by VEAO Simulations': True,
            'F-86F Sabre AI by Eagle Dynamics': True
        }

    def load_file(self, filename: str):
        """Load a mission file (.miz) file, replacing all current data.

        Args:
            filename: path to the mission(.miz) file.

        Returns:
            bool: True if everything loaded correctly

        Raises:
            RuntimeError: if an unknown value is encountered
        """
        self.filename = filename
        self.current_unit_id = 0
        self.current_group_id = 0
        self.current_dict_id = 0
        mission_dict = {}
        options_dict = {}
        warehouse_dict = {}
        dictionary_dict = {}

        def loaddict(fname, mizfile):
            with mizfile.open(fname) as mfile:
                data = mfile.read()
                data = data.decode()
                return lua.loads(data)

        with zipfile.ZipFile(filename, 'r') as miz:
            mission_dict = loaddict('mission', miz)
            if mission_dict["mission"]["version"] < 9:
                print("Mission file is using an old format, be aware!", file=sys.stderr)
            options_dict = loaddict('options', miz)
            warehouse_dict = loaddict('warehouses', miz)
            dictionary_dict = loaddict('l10n/DEFAULT/dictionary', miz)
            if 'l10n/DEFAULT/mapResource' in miz.namelist():
                mapresource_dict = loaddict('l10n/DEFAULT/mapResource', miz)
                self.map_resource.load_from_dict(mapresource_dict, miz)

        imp_mission = mission_dict["mission"]

        # import translations
        self.translation = Translation(self)
        translation_dict = dictionary_dict["dictionary"]
        for sid in translation_dict:
            self.translation.set_string(sid, translation_dict[sid], 'DEFAULT')

        self.current_dict_id = imp_mission["maxDictId"]

        # print(self.translation)

        # setup terrain
        if imp_mission["theatre"] == 'Caucasus':
            self.terrain = Caucasus()
        elif imp_mission["theatre"] == 'Nevada':
            self.terrain = Nevada()
        else:
            raise RuntimeError("Unknown theatre: '{theatre}'".format(theatre=imp_mission["theatre"]))

        # import options
        self.options = Options()
        self.options.load_from_dict(options_dict["options"])

        # import warehouses
        self.warehouses = Warehouses(self.terrain)
        self.warehouses.load_dict(warehouse_dict["warehouses"])

        # import base values
        self._description_text = self.translation.get_string(imp_mission["descriptionText"])
        self._description_bluetask = self.translation.get_string(imp_mission["descriptionBlueTask"])
        self._description_redtask = self.translation.get_string(imp_mission["descriptionRedTask"])
        self._sortie = self.translation.get_string(imp_mission["sortie"])
        for pic in sorted(imp_mission["pictureFileNameR"]):
            self.pictureFileNameR.append(imp_mission["pictureFileNameR"][pic])
        for pic in sorted(imp_mission["pictureFileNameB"]):
            self.pictureFileNameB.append(imp_mission["pictureFileNameB"][pic])
        self.version = imp_mission["version"]
        self.currentKey = imp_mission["currentKey"]
        self.start_time = datetime.fromtimestamp(1306886400 + imp_mission["start_time"], timezone.utc)
        self.usedModules = imp_mission.get("usedModules", None)
        self.needModules = imp_mission["needModules"]

        # groundControl
        self.groundControl = GroundControl()
        self.groundControl.load_from_dict(imp_mission.get("groundControl"))

        # goals
        self.goals = Goals()
        self.goals.load_from_dict(imp_mission["goals"])

        self.init_script_file = imp_mission.get("initScriptFile")
        self.init_script = imp_mission.get("initScript")

        # triggers
        self.triggers = Triggers()
        self.triggers.load_from_dict(imp_mission["triggers"])

        # this will import trigrules and trig
        self.triggerrules.load_from_dict(self, imp_mission["trigrules"])

        # failures
        self.failures = imp_mission["failures"]  # TODO

        # forced options
        self.forced_options.load_from_dict(imp_mission["forcedOptions"])

        # map
        self.map.load_from_dict(imp_mission["map"])

        # weather
        self.season_from_start_time = False
        self.random_weather = False
        imp_weather = imp_mission["weather"]
        self.weather = weather.Weather(self.terrain)
        self.weather.load_from_dict(imp_weather)

        # import coalition with countries and units
        for col_name in ["blue", "red", "neutral"]:
            if col_name in imp_mission["coalition"]:
                self.coalition[col_name] = Coalition(col_name, imp_mission["coalition"][col_name]["bullseye"])
                self.coalition[col_name].load_from_dict(self, imp_mission["coalition"][col_name])

        return True

    def sortie_text(self) -> str:
        """Returns the mission sortie text.

        Returns:
            the mission sortie text
        """
        return str(self._sortie)

    def set_sortie_text(self, text: str):
        """Sets the mission sortie text.

        Args:
            text: text to set.
        """
        self._sortie.set(text)

    def description_text(self) -> str:
        """Returns the mission description text.

        Returns:
            the mission description text
        """
        return str(self._description_text)

    def set_description_text(self, text: str):
        """Sets the mission descsription text.

        Args:
            text: text to set.
        """
        self._description_text.set(text)

    def description_bluetask_text(self) -> str:
        """Returns the blue task description text.

        Returns:
            the blue task description text
        """
        return str(self._description_bluetask)

    def set_description_bluetask_text(self, text: str):
        """Sets the red coalitions task description text.

        Args:
            text: text to set.
        """
        self._description_bluetask.set(text)

    def description_redtask_text(self) -> str:
        """Returns the red task description text.

        Returns:
            the red task description text
        """
        return str(self._description_redtask)

    def set_description_redtask_text(self, text: str):
        """Sets the red coalitions task description text.

        Args:
            text: text to set.
        """
        self._description_redtask.set(text)

    def add_picture_red(self, filepath: str) -> str:
        """Adds a new briefing picture to the red coalition.

        Args:
            filepath: path to the image, jpg or bmp.

        Returns:
            the resource key of the picture
        """
        reskey = self.map_resource.add_resource_file(filepath)
        self.pictureFileNameR.append(reskey)

    def add_picture_blue(self, filepath: str) -> str:
        """Adds a new briefing picture to the blue coalition.

        Args:
            filepath: path to the image, jpg or bmp.

        Returns:
            the resource key of the picture
        """
        reskey = self.map_resource.add_resource_file(filepath)
        self.pictureFileNameB.append(reskey)
        return reskey

    def next_group_id(self):
        """Get the next free group id

        Returns:
            a new group id
        """
        self.current_group_id += 1
        return self.current_group_id

    def next_unit_id(self):
        """Get the next free unit id

        Returns:
            a new unit id
        """
        self.current_unit_id += 1
        return self.current_unit_id

    def next_dict_id(self):
        """Get the next free dictionary id

        Returns:
            a new dictionary id
        """
        self.current_dict_id += 1
        return self.current_dict_id

    def eplrs_for(self, group: str) -> Dict[int, int]:
        """Searches all vehicle eplrs using groups and writes them in a mapping

        Args:
            group: which group to look for eplrs task, ["helicopter", "plane", "vehicle"]

        Returns:
            a dict mapping groups to used eplrs id
        """
        eplrs_map = {}
        for col in self.coalition:
            for c_name, country in self.coalition[col].countries.items():
                search_group = []
                if group == "helicopter":
                    search_group = country.helicopter_group
                elif group == "plane":
                    search_group = country.plane_group
                elif group == "vehicle":
                    search_group = country.vehicle_group
                for grp in search_group:
                    if grp.points:
                        eplrs = grp.points[0].find_task(task.EPLRS)
                        if eplrs:
                            eplrs_map[grp.id] = eplrs.eplrs
        return eplrs_map

    def next_eplrs(self, group_type: str) -> int:
        """Get next eplrs for the given group type.

        Args:
            group_type: one of "vehicle", "helicopter" or "plane"

        Returns:
            int: the next eplrs id to use
        """
        eplrs_usage = self.eplrs_for(group_type)
        eplrs_id = 1
        for ep_id in [eplrs_usage[x] for x in eplrs_usage]:
            if ep_id != eplrs_id:
                return eplrs_id
            eplrs_id += 1
        return eplrs_id

    def string(self, s, lang='DEFAULT'):
        """Create a new String() object for translation

        Args:
            s: string for lang
            lang: language for s

        Returns:
            String: A new String() object for string s
        """
        return self.translation.create_string(s, lang)

    def static(self, name, _type: unittype.UnitType) -> Static:
        """Creates a plain static object to be added to a group

        Args:
            name: of the static object
            _type(StaticType): type of the static

        Returns:
            Static: a new static object
        """
        return Static(self.next_unit_id(), self.string(name), _type)

    def static_group(self, country, name, _type: unittype.UnitType, position: mapping.Point,
                     heading=0, hidden=False, dead=False):
        """Add a static group with 1 static object.

        Args:
            country(Country): the object belongs too
            name: name of the group
            _type: what kind of object
            position(dcs.mapping.Point): where to place the object
            heading: of the object
            hidden: should the object be hidden on the map
            dead: should the object be rendered as dead

        Returns:
            StaticGroup: the new static group
        """
        sg = unitgroup.StaticGroup(self.next_group_id(), self.string(name))

        s = self.static(name + " object", _type)
        s.position = copy.copy(position)
        s.heading = heading
        sg.add_unit(s)

        sg.hidden = hidden
        sg.dead = dead

        sp = StaticPoint()
        sp.position = s.position
        sg.add_point(sp)

        country.add_static_group(sg)
        return sg

    def vehicle(self, name, _type: unittype.VehicleType) -> Vehicle:
        """Creates a plain vehicle unit to be added to a group

        Args:
            name: of the vehicle
            _type: vehicle type

        Returns:
            Vehicle: a new vehicle unit.
        """
        if not issubclass(_type, unittype.VehicleType):
            raise TypeError("_type not a unittype.VehicleType class: " + repr(_type))
        return Vehicle(self.next_unit_id(), self.string(name), _type.id)

    def vehicle_group(self, country, name, _type: unittype.VehicleType, position: mapping.Point,
                      heading=0, group_size=1,
                      formation=unitgroup.VehicleGroup.Formation.Line,
                      move_formation: PointAction=PointAction.OffRoad) -> unitgroup.VehicleGroup:
        """Adds a new vehicle group to the given country.

        Args:
                country(Country):which the vehicle group will belong too
            name: of the vehicle group
            _type: type of vehicle
            position: :py:class:`dcs.mapping.Point` where the new group will be placed
            heading: initial heading of the group, only used if no additional waypoints
            group_size: how many vehicles to add
            formation: formation in which the group should be placed
            move_formation: formation the group should use for moving

        Returns:
            VehicleGroup: the new vehicle group object
        """
        vg = unitgroup.VehicleGroup(self.next_group_id(), self.string(name))

        for i in range(1, group_size + 1):
            v = self.vehicle(name + " Unit #{nr}".format(nr=i), _type)
            v.position.x = position.x
            v.position.y = position.y + (i - 1) * 20
            v.heading = heading
            vg.add_unit(v)

        wp = vg.add_waypoint(vg.units[0].position, move_formation, 0)
        wp.ETA_locked = True
        if _type.eplrs:
            wp.tasks.append(task.EPLRS(self.next_eplrs("vehicle")))

        vg.formation(formation, heading)

        country.add_vehicle_group(vg)
        return vg

    def vehicle_group_platoon(self, country, name,
                              types: List[unittype.VehicleType],
                              position: mapping.Point,
                              heading=0,
                              formation=unitgroup.VehicleGroup.Formation.Line,
                              move_formation: PointAction=PointAction.OffRoad) -> unitgroup.VehicleGroup:
        """Adds a new vehicle group to the given country and given vehicle types.

        Args:
                country(Country):which the vehicle group will belong too
            name: of the vehicle group
            types: a list of vehicle types that will be used the units
            position: :py:class:`dcs.mapping.Point` where the new group will be placed
            heading: initial heading of the group, only used if no additional waypoints
            formation: formation in which the group should be placed
            move_formation: formation the group should use for moving

        Returns:
            VehicleGroup: the new vehicle group object
        """
        vg = unitgroup.VehicleGroup(self.next_group_id(), self.string(name))

        eplrs = False
        for i in range(0, len(types)):
            utype = types[i]
            v = self.vehicle(name + " Unit #{nr}".format(nr=i + 1), utype)
            v.position.x = position.x
            v.position.y = position.y + i * 20
            v.heading = heading
            if utype.eplrs:
                eplrs = True
            vg.add_unit(v)

        wp = vg.add_waypoint(vg.units[0].position, move_formation, 0)
        wp.ETA_locked = True
        if eplrs:
            wp.tasks.append(task.EPLRS(self.next_eplrs("vehicle")))

        vg.formation(formation, heading)

        country.add_vehicle_group(vg)
        return vg

    def ship(self, name, _type: unittype.ShipType) -> Ship:
        """Creates a plain ship unit to be added to a group

        Args:
            name: of the ship
            _type: ship type

        Returns:
            Ship: a new ship unit.
        """
        return Ship(self.next_unit_id(), self.string(name), _type)

    def ship_group(self, country, name, _type: unittype.ShipType,
                   position: mapping.Point, heading=0, group_size=1) -> unitgroup.ShipGroup:
        """Adds a ship group to the given country.

        Args:
            country(Country): which the ship group will belong too
            name: of the ship group
            _type: which kind of ship to add
            position(dcs.mapping.Point): where the new group will be placed
            heading: initial heading of the group, only used if no additional waypoints
            group_size: how many ships of _type

        Returns:
            ShipGroup: the new ship group object
        """
        sg = unitgroup.ShipGroup(self.next_group_id(), self.string(name))

        for i in range(1, group_size + 1):
            v = self.ship(name + " Unit #{nr}".format(nr=i), _type)
            v.position.x = position.x
            v.position.y = position.y + (i - 1) * 20
            v.heading = heading
            sg.add_unit(v)

        wp = sg.add_waypoint(position, 20)
        wp.ETA_locked = True

        country.add_ship_group(sg)
        return sg

    def plane_group(self, name) -> unitgroup.PlaneGroup:
        """This creates a plain plane group without any units or starting points.

        This method is a advanced interface method not intended for simple use.
        For adding full featured plane group see

         * :py:meth:`flight_group`
         * :py:meth:`flight_group_inflight`
         * :py:meth:`flight_group_from_airport`

        Args:
            name: Group name

        Returns:
            PlaneGroup: A new :py:class:`dcs.unitgroup.PlaneGroup`
        """
        return unitgroup.PlaneGroup(self.next_group_id(), self.string(name))

    def plane(self, name, _type: planes.PlaneType, country: Country):
        """Creates a new plane unit.

        This method is a advanced interface method not intended for simple usage.
        For adding full a featured plane group see

         * :py:meth:`flight_group`
         * :py:meth:`flight_group_inflight`
         * :py:meth:`flight_group_from_airport`

        Args:
            name: unit name
            _type: type of the plane
            country(Country): the plane belongs, needed for default liveries

        Return:
            Plane: A new :py:class:`dcs.unit.Plane`
        """
        return Plane(self.next_unit_id(), self.string(name), _type, country)

    def helicopter(self, name, _type: helicopters.HelicopterType, country: Country):
        """Creates a new helicopter unit.

        This method is a advanced interface method not intended for simple usage.
        For adding full a featured helicopter group see

         * :py:meth:`flight_group`
         * :py:meth:`flight_group_inflight`
         * :py:meth:`flight_group_from_airport`

        Args:
            name: unit name
            _type: type of the helicopter
            country(Country): the helicopter belongs, needed for default liveries

        Returns:
            Helicopter: A new :py:class:`dcs.unit.Helicopter`
        """
        return Helicopter(self.next_unit_id(), self.string(name), _type, country)

    def aircraft(self, name, _type: unittype.FlyingType, country: Country) -> Union[Plane, Helicopter]:
        """Creates a new plane or helicopter unit, depending on the _type.

        This method is a advanced interface method not intended for simple usage.
        For adding full a featured plane/helicopter group see

         * :py:meth:`flight_group`
         * :py:meth:`flight_group_inflight`
         * :py:meth:`flight_group_from_airport`

        Args:
            name: unit name
            _type: type of the aircraft
            country(Country): the aircraft belongs, needed for default liveries

        Returns:
            Helicopter: A new :py:class:`dcs.unit.Plane` or :py:class:`dcs.unit.Helicopter`
        """
        if _type.helicopter:
            return Helicopter(self.next_unit_id(), self.string(name), _type, country)
        return Plane(self.next_unit_id(), self.string(name), _type, country)

    def helicopter_group(self, name) -> unitgroup.HelicopterGroup:
        """Creates a plain helicopter group without any units or starting points.

        This method is a advanced interface method not intended for simple usage.
        For adding a full featured helicopter group see

         * :py:meth:`flight_group`
         * :py:meth:`flight_group_inflight`
         * :py:meth:`flight_group_from_airport`

        Args:
             name: Group name

        Returns:
            HelicopterGroup: A new :py:class:`dcs.unitgroup.HelicopterGroup`
        """
        return unitgroup.HelicopterGroup(self.next_group_id(), self.string(name))

    @classmethod
    def _assign_callsign(cls, _country, group):
        callsign_name = None
        category = "Air" if group.units[0].unit_type.category == "Interceptor" else group.units[0].unit_type.category
        if category in _country.callsign:
            callsign_name = _country.next_callsign_category(category)

        i = 1
        for unit in group.units:
            if category in _country.callsign:
                unit.callsign_dict["name"] = callsign_name + str(1) + str(i)
                unit.callsign_dict[3] = i
            else:
                unit.callsign = _country.next_callsign_id()
            i += 1

    @staticmethod
    def _load_tasks(mp: MovingPoint, maintask: task.MainTask):
        for t in maintask.perform_task:
            ptask = t()
            ptask.auto = True
            mp.tasks.append(ptask)
        return mp

    def _flying_group_from_airport(self, _country, group: unitgroup.FlyingGroup,
                                   maintask: task.MainTask,
                                   airport: Airport,
                                   start_type: StartType=StartType.Cold,
                                   parking_slots: List[ParkingSlot] = None) -> unitgroup.FlyingGroup:

        for unit in group.units:
            spos = airport.position
            if start_type != StartType.Runway:
                parking_slot = parking_slots.pop(0) if parking_slots else airport.free_parking_slot(
                    unit.unit_type)
                if parking_slot is None:
                    raise NoParkingSlotError("No free parking slot at " + airport.name)
                spos = parking_slot.position
                unit.set_parking(parking_slot)
            unit.position = copy.copy(spos)

        if start_type == StartType.Runway:
            airport.occupy_runway(group)

        group.load_task_default_loadout(maintask)

        self._assign_callsign(_country, group)

        point_start_type_map = {
            StartType.Cold: ("TakeOffParking", PointAction.FromParkingArea),
            StartType.Warm: ("TakeOffParkingHot", PointAction.FromParkingAreaHot),
            StartType.Runway: ("TakeOff", PointAction.FromRunway)
        }
        mp = MovingPoint()
        mp.type = point_start_type_map[start_type][0]
        mp.action = point_start_type_map[start_type][1]
        mp.position = copy.copy(group.units[0].position)
        mp.airdrome_id = airport.id
        mp.alt = group.units[0].alt
        mp.properties = PointProperties()
        Mission._load_tasks(mp, maintask)
        first_unit = group.units[0]
        if first_unit.unit_type.eplrs:
            mp.tasks.append(task.EPLRS(self.next_eplrs("helicopter" if first_unit.unit_type.helicopter else "plane")))

        group.add_point(mp)

        return group

    def _flying_group_inflight(self, _country, group: unitgroup.FlyingGroup,
                               maintask: task.MainTask, altitude, speed) -> unitgroup.FlyingGroup:

        i = 0
        for unit in group.units:
            unit.alt = altitude
            unit.position.x += i * 10
            unit.speed = speed / 3.6
            i += 1

        self._assign_callsign(_country, group)

        group.load_task_default_loadout(maintask)

        mp = MovingPoint()
        mp.type = "Turning Point"
        mp.action = PointAction.TurningPoint
        mp.position = copy.copy(group.units[0].position)
        mp.alt = altitude
        mp.speed = speed / 3.6
        mp.properties = PointProperties()

        Mission._load_tasks(mp, maintask)
        first_unit = group.units[0]
        if first_unit.unit_type.eplrs:
            mp.tasks.append(task.EPLRS(self.next_eplrs("helicopter" if first_unit.unit_type.helicopter else "plane")))

        group.add_point(mp)

        return group

    def flight_group_inflight(self,
                              country,
                              name: str,
                              aircraft_type: unittype.FlyingType,
                              position: mapping.Point,
                              altitude: int,
                              speed=None,
                              maintask: Optional[task.MainTask] = None,
                              group_size: int=1
                              ) -> Union[unitgroup.PlaneGroup, unitgroup.HelicopterGroup]:
        """Add a new Plane/Helicopter group inflight.

        The type of the resulting group depends on the given aircraft_type.

        Args:
            country(Country): the new group will belong to
            name: of the new group
            aircraft_type(FlyingType): type of all units in the group
            position(dcs.mapping.Point): where the new group will be placed
            altitude: of the new group
            speed: of the new group, if none a default will be picked
            maintask(MainTask): if none the default task for the aircraft_type wil be used
            group_size: number of units in the group(maximum 4 or 1 for certain types)

        Returns:
            FlyingGroup: a new :py:class:`dcs.unitgroup.PlaneGroup` or :py:class:`dcs.unitgroup.HelicopterGroup`
        """
        if maintask is None:
            maintask = aircraft_type.task_default

        if aircraft_type.helicopter:
            ag = self.helicopter_group(name)
            speed = speed if speed else 200
        else:
            ag = self.plane_group(name)
            speed = speed if speed else 600
        ag.task = maintask.name
        group_size = min(group_size, aircraft_type.group_size_max)

        for i in range(1, group_size + 1):
            p = self.aircraft(name + " Pilot #{nr}".format(nr=i), aircraft_type, country)
            p.position = copy.copy(position)
            p.fuel *= 0.9
            ag.add_unit(p)

        country.add_aircraft_group(self._flying_group_inflight(country, ag, maintask, altitude, speed))
        return ag

    def flight_group_from_airport(self,
                                  country: Country,
                                  name,
                                  aircraft_type: unittype.FlyingType,
                                  airport: Airport,
                                  maintask: task.MainTask = None,
                                  start_type: StartType=StartType.Cold,
                                  group_size=1,
                                  parking_slots: List[ParkingSlot] = None) -> \
            Union[unitgroup.PlaneGroup, unitgroup.HelicopterGroup]:
        """Add a new Plane/Helicopter group at the given airport.

        Runway, warm/cold start depends on the given start_type.

        Args:
            country(Country): Country object the plane group belongs to
            name: Name of the aircraft group
            maintask(MainTask): Task of the aircraft group
            aircraft_type(FlyingType): FlyingType class that describes the aircraft_type
            airport(Airport): Airport object on which to spawn the helicopter
            start_type(StartType): Start from runway, cold or warm parking position
            parking_slots: List of parking slots to use for aircrafts
            group_size: number of units in the group(maximum 4 or 1 for certain types)

        Returns:
            FlyingGroup: a new :py:class:`dcs.unitgroup.PlaneGroup` or :py:class:`dcs.unitgroup.HelicopterGroup`
        """
        if maintask is None:
            maintask = aircraft_type.task_default

        ag = self.helicopter_group(name) if aircraft_type.helicopter else self.plane_group(name)
        ag.task = maintask.name
        group_size = min(group_size, aircraft_type.group_size_max)

        for i in range(1, group_size + 1):
            p = self.aircraft(name + " Pilot #{nr}".format(nr=i), aircraft_type, country)
            ag.add_unit(p)

        country.add_aircraft_group(
            self._flying_group_from_airport(country, ag, maintask, airport, start_type, parking_slots))
        return ag

    def flight_group_from_unit(self,
                               country: Country,
                               name,
                               aircraft_type: unittype.FlyingType,
                               carrier_unit: Union[Ship, Static],
                               maintask: task.MainTask = None,
                               start_type: StartType = StartType.Cold,
                               group_size=1) -> Union[unitgroup.PlaneGroup, unitgroup.HelicopterGroup]:
        """Add a new Plane/Helicopter group at the given FARP or carrier unit.

        Args:
            country(Country): Country object the plane group belongs to
            name: Name of the aircraft group
            maintask(MainTask): Task of the aircraft group
            aircraft_type(FlyingType): FlyingType class that describes the aircraft_type
            carrier_unit(Unit): Group(Ship, FARP) on which to spawn
            start_type(StartType): Start from runway, cold or warm parking position, ignored for now
            group_size: number of units in the group(maximum 4 or 1 for certain types)

        Returns:
            FlyingGroup: a new :py:class:`dcs.unitgroup.PlaneGroup` or :py:class:`dcs.unitgroup.HelicopterGroup`
        """
        if maintask is None:
            maintask = aircraft_type.task_default

        ag = self.helicopter_group(name) if aircraft_type.helicopter else self.plane_group(name)
        ag.task = maintask.name
        group_size = min(group_size, aircraft_type.group_size_max)

        for i in range(1, group_size + 1):
            p = self.aircraft(name + " Pilot #{nr}".format(nr=i), aircraft_type, country)
            ag.add_unit(p)

        ag.units[0].position = copy.copy(carrier_unit.position)
        ag.formation_rectangle(carrier_unit.heading, 10)

        ag.load_task_default_loadout(maintask)

        self._assign_callsign(country, ag)

        point_start_type_map = {
            StartType.Cold: ("TakeOffParking", PointAction.FromParkingArea),
            StartType.Warm: ("TakeOffParkingHot", PointAction.FromParkingAreaHot),
            StartType.Runway: ("TakeOff", PointAction.FromRunway)
        }
        mp = MovingPoint()
        mp.type = point_start_type_map[start_type][0]
        mp.action = point_start_type_map[start_type][1]
        mp.position = copy.copy(ag.units[0].position)
        mp.helipad_id = carrier_unit.id
        mp.alt = ag.units[0].alt
        mp.properties = PointProperties()
        Mission._load_tasks(mp, maintask)
        first_unit = ag.units[0]
        if first_unit.unit_type.eplrs:
            mp.tasks.append(task.EPLRS(self.next_eplrs("helicopter" if first_unit.unit_type.helicopter else "plane")))

        ag.add_point(mp)

        country.add_aircraft_group(ag)
        return ag

    def flight_group(self,
                     country: Country,
                     name: str,
                     aircraft_type: unittype.FlyingType,
                     airport: Optional[Airport],
                     position: Optional[mapping.Point],
                     altitude=3000,
                     speed=500,
                     maintask: Optional[task.MainTask]=None,
                     start_type: StartType=StartType.Runway,
                     group_size=1
                     ) -> unitgroup.FlyingGroup:
        """This is wrapper around flight_group_inflight and flight_group_from_airport.

        Depending on the airport parameter a flight group will added inflight or on an airport.

        Args:
            country(Country): Country object the plane group belongs to
            name: Name of the aircraft group
            aircraft_type(FlyingType): FlyingType class that describes the aircraft_type
            airport(Airport): Airport object on which to spawn the helicopter
            position(dcs.mapping.Point): where the new group will be placed, if inflight
            altitude: initial altitude of the group if inflight
            speed: initial speed of the group if inflight
            maintask(MainTask): Task of the aircraft group
            start_type(StartType): Start from runway, cold or warm parking position
            group_size: number of units in the group(maximum 4 or 1 for certain types)

        Returns:
            FlyingGroup: a new :py:class:`dcs.unitgroup.PlaneGroup` or :py:class:`dcs.unitgroup.HelicopterGroup`
        """
        if airport:
            fg = self.flight_group_from_airport(country, name, aircraft_type,
                                                airport, maintask, start_type, group_size)
        else:
            fg = self.flight_group_inflight(country, name, aircraft_type,
                                            position, altitude, speed, maintask, group_size)

        return fg

    def awacs_flight(self,
                     country: Country,
                     name: str,
                     plane_type: planes.PlaneType,
                     airport: Optional[Airport],
                     position: mapping.Point,
                     race_distance=30 * 1000,
                     heading=90,
                     altitude=4500,
                     speed=550,
                     start_type: StartType=StartType.Cold,
                     frequency=140) -> unitgroup.PlaneGroup:
        """Add an AWACS flight group.

        This is simple way to add an AWACS flight group to your mission.
        It needs an initial orbit point, race distance and heading from this point.

        If an airport is given the AWACS flight will start from there otherwise,
        it will placed 2 km in front of the reference position.

        Args:
            country(Country): Country object the awacs group belongs to
            name: of the AWACS flight
            plane_type(PlaneType): AWACS plane type. e.g E_3A
            airport(Airport): starting airport, use None if you want it to spawn inflight
            position(dcs.mapping.Point): reference point for the race-track
            race_distance: distance for the race-track pattern
            heading: direction from the referene position
            altitude: of the AWACS race-track
            speed: of the AWACS flight
            start_type(StartType): of the flight if starts from airport
            frequency: VHF-AM frequencey in mhz

        Returns:
            PlaneGroup: the created AWACS flight group
        """
        if airport:
            awacs = self.flight_group_from_airport(country, name, plane_type, airport, task.AWACS, start_type)
            wp = awacs.add_runway_waypoint(airport)
        else:
            p = position.point_from_heading((heading + 180) % 360, 2000)
            awacs = self.flight_group_inflight(country, name, plane_type, p, altitude, speed, task.AWACS)
            p = position.point_from_heading(heading + 180, 1000)
            wp = awacs.add_waypoint(p, altitude, speed)

        wp.tasks.append(task.SetFrequencyCommand(frequency))

        wp = awacs.add_waypoint(position, altitude, speed)
        wp.tasks.append(task.OrbitAction(altitude, speed, task.OrbitAction.OrbitPattern.RaceTrack))

        p = position.point_from_heading(heading, race_distance)
        awacs.add_waypoint(p, altitude, speed)

        return awacs

    def refuel_flight(self,
                      country,
                      name: str,
                      plane_type: planes.PlaneType,
                      airport: Optional[Airport],
                      position: mapping.Point,
                      race_distance=30 * 1000,
                      heading=90,
                      altitude=4500,
                      speed=407,
                      start_type: StartType=StartType.Cold,
                      frequency=140,
                      tacanchannel="10X") -> unitgroup.PlaneGroup:
        """Add an refuel flight group.

        This is simple way to add an refuel flight group to your mission.
        It needs an initial orbit point, race distance and heading from this point.

        If an airport is given the refuel flight will start from there otherwise,
        it will placed 2 km in front of the reference position.

        Args:
            country(Country): Country object the awacs group belongs to
            name: of the refuel flight
            plane_type(PlaneType): refuel plane type. e.g KC_135
            airport(Airport): starting airport, use None if you want it to spawn inflight
            position(dcs.mapping.Point): reference point for the race-track
            race_distance: distance for the race-track pattern
            heading: direction from the referene position
            altitude: of the refuel race-track
            speed: of the refuel flight
            start_type(StartType): of the flight if starts from airport
            frequency: VHF-AM frequencey in mhz
            tacanchannel: if the PlaneType supports tacan this channel will be set.

        Returns:
            PlaneGroup: the created refuel flight group
        """
        if airport:
            tanker = self.flight_group_from_airport(country, name, plane_type, airport,
                                                    task.Refueling, start_type=start_type)
            wp = tanker.add_runway_waypoint(airport)
        else:
            p = position.point_from_heading((heading + 180) % 360, 2000)
            tanker = self.flight_group_inflight(country, name, plane_type, p, altitude, speed, task.Refueling)
            p = position.point_from_heading(heading + 180, 1000)
            wp = tanker.add_waypoint(p, altitude, speed)

        wp.tasks.append(task.SetFrequencyCommand(frequency))

        if plane_type.tacan:
            channel = int(tacanchannel[:-1])
            modechannel = tacanchannel[-1]
            tanker.points[0].tasks.append(task.ActivateBeaconCommand(channel, modechannel))

        wp = tanker.add_waypoint(position, altitude, speed)
        wp.tasks.append(task.OrbitAction(altitude, speed, task.OrbitAction.OrbitPattern.RaceTrack))

        p = position.point_from_heading(heading, race_distance)
        tanker.add_waypoint(p, altitude, speed)

        return tanker

    def escort_flight(self,
                      country,
                      name: str,
                      escort_type: planes.PlaneType,
                      airport: Optional[Airport],
                      group_to_escort: unitgroup.FlyingGroup,
                      start_type: StartType=StartType.Cold,
                      group_size=2):
        """Add an escort flight group to the mission.

        An escort flight is a flight group that will use the :py:class:`dcs.task.EscortTaskAction`
        to escort another flight group.

        If no airport is given, the escort flight will spawn near the group to escort.

        Args:
            country(Country): the escort flight belongs too
            name: of the flight group
            escort_type(PlaneType): PlaneType for the escort task
            airport(Airport): starting airport, use None if you want it to spawn inflight
            group_to_escort: id of the group to escort
            start_type(StartType): of the flight if starts from airport
            group_size: how many planes should be in the escort flight

        Returns:
            PlaneGroup: the created escort group
        """
        second_point_group = group_to_escort.points[1]
        if airport:
            eg = self.flight_group_from_airport(
                country, name, escort_type, airport, task.Escort, start_type=start_type, group_size=group_size
            )
            eg.add_runway_waypoint(airport)
        else:
            eg = self.flight_group_inflight(
                country, name, escort_type,
                mapping.Point(group_to_escort.points[0].position.x - 10 * 1000, group_to_escort.points[0].position.y),
                second_point_group.alt + 200,
                maintask=task.Escort,
                group_size=group_size
            )

        eg.add_waypoint(second_point_group.position, second_point_group.alt)
        eg.points[0].tasks[0] = task.EscortTaskAction(group_to_escort.id, lastwpt=len(group_to_escort.points))

        return eg

    def patrol_flight(self,
                      country,
                      name: str,
                      patrol_type: planes.PlaneType,
                      airport: Optional[Airport],
                      pos1,
                      pos2,
                      start_type: StartType=StartType.Cold,
                      speed=600,
                      altitude=4000,
                      max_engage_distance=60*1000,
                      group_size=2):
        """Add an patrol flight group to the mission.

        A patrol flight is a flight group that will fly a orbit between 2 given points and
        will engage any incoming air threats within max_engage_distance.

        If no airport is given, the patrol flight will spawn near the first patrol point(pos1).

        Args:
            country(Country): the flight belongs too
            name: name of the patrol flight
            patrol_type(PlaneType): PlaneType for the patrol flight
            airport(Airport): starting airport, use None if you want it to spawn inflight
            pos1(dcs.mapping.Point): first orbit waypoint
            pos2(dcs.mapping.Point): second orbit waypoint
            start_type(StartType): of the flight if starts from airport
            speed: orbit speed
            altitude: initial altitude and orbit altitude
            max_engage_distance: the distance in KM the patrol flight will respond to enemy threats
            group_size: how many planes should be in the flight group

        Returns:
            PlaneGroup: the created patrol group
        """
        if airport:
            eg = self.flight_group_from_airport(
                country, name, patrol_type, airport, maintask=task.CAP, start_type=start_type, group_size=group_size
            )
            eg.add_runway_waypoint(airport)
        else:
            eg = self.flight_group_inflight(
                country, name, patrol_type,
                mapping.Point(pos1.x - 10 * 1000, pos1.y),
                altitude,
                maintask=task.CAP,
                group_size=group_size
            )
        eg.points[0].tasks[0] = task.EngageTargets(max_engage_distance, [task.Targets.All.Air])
        wp = eg.add_waypoint(pos1, altitude, speed)
        wp.tasks.append(task.OrbitAction(altitude, speed, task.OrbitAction.OrbitPattern.RaceTrack))
        eg.add_waypoint(pos2, altitude, speed)

        return eg

    def country(self, name):
        """Returns the country object for the mission by the given string

        Args:
            name: string representation of the country

        Returns:
            Country: the object of the country, None if not found.
        """
        for k in self.coalition:
            c = self.coalition[k].country(name)
            if c:
                return c
        return None

    def find_group(self, group_name, search="exact") -> Optional[unitgroup.Group]:
        """Searches a group with the given name.

        Args:
            group_name: part or exact name of the group
            search: search mode to use

                      * 'exact': whole name must match
                      * 'match': part of the name must match

        Returns:
            Group: the group found, otherwise None
        """
        for k in self.coalition:
            g = self.coalition[k].find_group(group_name, search)
            if g:
                return g
        return None

    def is_red(self, country: Country) -> bool:
        """Checks if the given country object is part o the red coalition.

        Args:
            country(Country): object to check

        Returns:
            bool: True if it is part of the red coalition, else False.
        """
        return country.name in self.coalition["red"].countries

    def is_blue(self, country: Country) -> bool:
        """Checks if the given country object is part o the blue coalition.

        Args:
                country(Country):object to check

        Returns:
            bool: True if it is part of the blue coalition, else False.
        """
        return country.name in self.coalition["blue"].countries

    def random_date(self):
        """Sets a random date for the mission"""
        self.start_time = datetime.fromtimestamp(1306886400 + 43200, timezone.utc)  # 01-06-2011 12:00:00 UTC
        self.start_time += timedelta(days=random.randrange(0, 365))

    def random_daytime(self, period):
        self.start_time = datetime(
            year=self.start_time.year,
            month=self.start_time.month,
            day=self.start_time.day,
            tzinfo=self.start_time.tzinfo)
        daytime_map = {
            "day": timedelta(minutes=random.randrange(420, 1140)),
            "night": timedelta(minutes=random.randrange(-120, 240)),
            "dusk": timedelta(minutes=random.randrange(960, 1100)),
            "dawn": timedelta(minutes=random.randrange(240, 480)),
            "noon": timedelta(hours=random.randrange(600, 840))
        }
        if period == "random":
            period = random.choice(list(daytime_map.keys()))
        self.start_time += daytime_map[period]

    def stats(self) -> Dict:
        """Gather some mission stats.

        This method counts up the different group types and used units
        and returns them as easy to print dict.

        Returns:
            dict containing various group and unit counts.
        """
        d = {
            "red": {},
            "blue": {},
            "unit_count": 0,
            "count": 0
        }

        def count_group(field, group):
            d[col_name]["count"] += len(group)
            d[col_name][field]["count"] += len(group)
            for g in group:
                for u in g.units:
                    _unit = d[col_name][field]["units"].get(u.type, 0)
                    d[col_name][field]["units"][u.type] = _unit + 1
                    d[col_name]["unit_count"] += 1
            d[col_name][field]["unit_count"] = sum(d[col_name][field]["units"].values())

        for col_name in ["red", "blue"]:
            d[col_name]["unit_count"] = 0
            d[col_name]["count"] = 0
            col = self.coalition[col_name]
            d[col_name]["plane_groups"] = {"count": 0, "units": {}}
            d[col_name]["helicopter_groups"] = {"count": 0, "units": {}}
            d[col_name]["vehicle_groups"] = {"count": 0, "units": {}}
            d[col_name]["ship_groups"] = {"count": 0, "units": {}}
            for k, v in col.countries.items():
                count_group("plane_groups", v.plane_group)
                count_group("helicopter_groups", v.helicopter_group)
                count_group("vehicle_groups", v.vehicle_group)
                count_group("ship_groups", v.ship_group)
            d["unit_count"] += d[col_name]["unit_count"]
            d["count"] += d[col_name]["count"]

        # import pprint
        # pp = pprint.PrettyPrinter(indent=2)
        # pp.pprint(d)
        return d

    def print_stats(self, d):
        """Print the given mission stats to standard output.

        Args:
            d: stats dict to print, :py:meth:`dcs.mission.Mission.stats`
        """
        print("Mission Statistics")
        print(self.start_time.strftime("%d. %b %H:%M:%S"))
        print("-"*60)
        output = {"red": [], "blue": []}
        for x in ["Blue", "Red"]:
            low = x.lower()
            output[low].append("{group:<15s} groups units".format(group=x))
            output[low].append("{group:<15s} {gc:6d} {u:5d}".format(
                group="Plane",
                gc=d[low]["plane_groups"]["count"], u=d[low]["plane_groups"]["unit_count"]))
            output[low].append("{group:<15s} {gc:6d} {u:5d}".format(
                group="Helicopter",
                gc=d[low]["helicopter_groups"]["count"], u=d[low]["helicopter_groups"]["unit_count"]))
            output[low].append("{group:<15s} {gc:6d} {u:5d}".format(
                group="Vehicle",
                gc=d[low]["vehicle_groups"]["count"], u=d[low]["vehicle_groups"]["unit_count"]))
            output[low].append("{group:<15s} {gc:6d} {u:5d}".format(
                group="Ship",
                gc=d[low]["ship_groups"]["count"], u=d[low]["ship_groups"]["unit_count"]))
            output[low].append("-"*28)
            output[low].append("{group:<15s} {gc:6d} {u:5d}".format(group="Sum", gc=d[low]["count"], u=d[low]["unit_count"]))

        # merge tables
        for i in range(0, len(output["blue"])):
            print(output["blue"][i], "  ", output["red"][i])
        print("Total {g} groups with {u} units".format(g=d["count"], u=d["unit_count"]))

    def reload(self):
        """Reloads the current loaded file

        Raises:
            RuntimeError: if there is currently no file loaded.
        """
        if self.filename:
            return self.load_file(self.filename)
        raise RuntimeError("Currently no file loaded to reload.")

    def save(self, filename=None):
        """Save the current Mission object to the given file.

        Args:
            filename: filepath to save the Mission object
            show_stats(bool): if True print mission stats to standard out.
        """
        filename = self.filename if filename is None else filename
        if not filename:
            raise RuntimeError("No filename given.")
        self.filename = filename  # store filename

        with zipfile.ZipFile(filename, 'w', compression=zipfile.ZIP_DEFLATED) as zipf:
            # options
            zipf.writestr('options', str(self.options))

            # warehouses
            zipf.writestr('warehouses', str(self.warehouses))

            # translation files
            dicttext = lua.dumps(self.translation.dict('DEFAULT'), "dictionary", 1)
            zipf.writestr('l10n/DEFAULT/dictionary', dicttext)

            mapresource = self.map_resource.store(zipf, 'DEFAULT')
            # print(mapresource)
            zipf.writestr('l10n/DEFAULT/mapResource', lua.dumps(mapresource, "mapResource", 1))

            zipf.writestr('mission', str(self))

        return True

    def dict(self):
        m = {
            "trig": self.triggerrules.trig(),
            "trigrules": self.triggerrules.trigrules(),
            "start_time": int(self.start_time.timestamp() - 1306886400)
        }
        if m["start_time"] < 0:
            raise RuntimeError("Mission start time is < 0.")
        m["date"] = {
            "Year": self.start_time.year,
            "Month": self.start_time.month,
            "Day": self.start_time.day
        }
        if self.season_from_start_time:
            self.weather.set_season_from_datetime(self.start_time)
        if self.random_weather:
            self.weather.random(self.start_time)
        m["groundControl"] = self.groundControl.dict()
        if self.usedModules is not None:
            m["usedModules"] = self.usedModules
        m["resourceCounter"] = self.resourceCounter
        m["triggers"] = self.triggers.dict()
        m["weather"] = self.weather.dict()
        m["theatre"] = self.terrain.name
        m["needModules"] = self.needModules
        m["map"] = self.map.dict()
        m["descriptionText"] = self._description_text.id
        m["pictureFileNameR"] = {}
        for i in range(0, len(self.pictureFileNameR)):
            m["pictureFileNameR"][i + 1] = self.pictureFileNameR[i]
        m["pictureFileNameB"] = {}
        for i in range(0, len(self.pictureFileNameB)):
            m["pictureFileNameB"][i + 1] = self.pictureFileNameB[i]
        m["descriptionBlueTask"] = self._description_bluetask.id
        m["descriptionRedTask"] = self._description_redtask.id
        if self.init_script_file is not None:
            m["initScriptFile"] = self.init_script_file
        if self.init_script is not None:
            m["initScript"] = self.init_script
        m["coalition"] = {}
        for col in self.coalition.keys():
            m["coalition"][col] = self.coalition[col].dict()
        col_blue = {self.coalition["blue"].country(x).id for x in self.coalition["blue"].countries.keys()}
        col_red = {self.coalition["red"].country(x).id for x in self.coalition["red"].countries.keys()}
        col_neutral = list(Mission._COUNTRY_IDS - col_blue - col_red)
        col_blue = list(col_blue)
        col_red = list(col_red)
        m["coalitions"] = {
            "neutral": {x + 1: col_neutral[x] for x in range(0, len(col_neutral))},
            "blue": {x + 1: col_blue[x] for x in range(0, len(col_blue))},
            "red": {x + 1: col_red[x] for x in range(0, len(col_red))}
        }
        m["sortie"] = self._sortie.id
        m["version"] = self.version
        m["goals"] = self.goals.dict()
        m["result"] = self.goals.generate_result()
        m["currentKey"] = self.currentKey
        m["maxDictId"] = self.current_dict_id
        m["forcedOptions"] = self.forced_options.dict()
        m["failures"] = self.failures

        return m

    def __str__(self):
        return lua.dumps(self.dict(), "mission", 1)

    def __repr__(self):
        rep = {"base": str(self), "options": self.options, "translation": self.translation}
        return repr(rep)


class MapResource:
    """MapResource is responsibly to manage all additional mission resource files.

    Mission resource files are briefing images, lua scripts, sounds files.

    Args:
        mission(Mission): the mission this MapResource belongs too, needed for dictionary ids
    """
    def __init__(self, mission: Mission):
        self.files = {}
        self.mission = mission

    def load_from_dict(self, _dict, zipf: zipfile.ZipFile, lang='DEFAULT'):
        _dict = _dict["mapResource"]

        for key in _dict:
            filename = _dict[key]
            extractedpath = zipf.extract('l10n/{lang}/{fn}'.format(lang=lang, fn=filename), tempfile.gettempdir())
            self.add_resource_file(extractedpath, lang, key)

    def add_resource_file(self, filepath, lang='DEFAULT', key=None):
        """Adds a file to the mission resource depot.

        Args:
            filepath: path to the file to add
            lang: language this file belongs too.
            key: should None, needed for loading

        Returns:
            resource key to use in scripts
        """
        abspath = os.path.abspath(filepath)
        resource_key = key if key else "ResKey_" + str(self.mission.next_dict_id())
        if lang not in self.files:
            self.files[lang] = {}
        self.files[lang][abspath] = {
            "path": abspath,
            "reskey": resource_key
        }
        return resource_key

    def store(self, zipf: zipfile.ZipFile, lang='DEFAULT'):
        d = {}
        if lang in self.files:
            for x in self.files[lang]:
                mr = self.files[lang][x]
                filepath = mr["path"]
                if os.path.isabs(filepath):
                    nameinzip = os.path.basename(filepath)
                    zipf.write(filepath, "l10n/{lang}/{name}".format(lang=lang, name=nameinzip))
                    d[mr["reskey"]] = nameinzip

        return d


class Options:
    """Should be a representation for the mission options file
    might be removed in the future.
    """
    def __init__(self):
        self.options = {}

    def load_from_dict(self, d):
        self.options = d

    def __str__(self):
        return lua.dumps(self.options, "options", 1)

    def __repr__(self):
        return repr(self.options)
